"""Tests for `scopus.AbstractRetrieval` module."""

from pybliometrics.scopus import AbstractRetrieval, init
from pybliometrics.scopus.abstract_retrieval import (
    Affiliation, AuthorGroup, Author, Chemical, Contributor, 
    Correspondence, Funding, ISSN, Reference, Sequencebank, Area
)

init()

# Base information
ab1 = AbstractRetrieval("2-s2.0-84930616647", view="FULL", refresh=30)
# Conference proceeding and no references
ab2 = AbstractRetrieval("2-s2.0-0029486824", view="FULL", refresh=30)
# Issuetitle and no affiliation
ab3 = AbstractRetrieval("2-s2.0-0001270077", view="FULL", refresh=30)
# Author group broken and author keywords
ab4 = AbstractRetrieval("2-s2.0-0000016206", view="FULL", refresh=30)
# ISBN
ab5 = AbstractRetrieval("2-s2.0-84919546381", view="FULL", refresh=30)
# Funding, sequencebanks, chemicals
ab6 = AbstractRetrieval("2-s2.0-85040230676", view="FULL", refresh=30)
# Contributor group
ab7 = AbstractRetrieval("2-s2.0-85050253030", view="FULL", refresh=30)
# REF view
ab8 = AbstractRetrieval("2-s2.0-84951753303", view="REF", refresh=30)
# Collaboration
ab9 = AbstractRetrieval("2-s2.0-85097473741", view="FULL", refresh=30)
# ENTITLED view
ar10 = AbstractRetrieval('10.1109/Multi-Temp.2019.8866947', view='ENTITLED', refresh=30)
# REF view without refs
ab11 = AbstractRetrieval('2-s2.0-85160105660', view="REF", refresh=30)
# FULL view with list of collaborations
ab12 = AbstractRetrieval('2-s2.0-85044008512', view='FULL', refresh=30)
# List of contributors in contributor group
ab13 = AbstractRetrieval("2-s2.0-85038825012", view="FULL", refresh=30)


def test_abstract():
    expected = 'In this paper we propose a Bayesian analysis of seasonal '\
        'unit roots in quarterly observed time series. Seasonal unit root '\
        'processes are useful to describe economic series with changing '\
        'seasonal fluctuations. A natural alternative model for similar '\
        'purposes contains deterministic seasonal mean shifts instead of '\
        'seasonal stochastic trends. This leads to analysing seasonal unit '\
        'roots in the presence of mean shifts using Bayesian techniques. '\
        'Our method is illustrated using several simulated and empirical data.'
    assert ab4.abstract == expected
    assert ab8.abstract is None


def test_affiliation():
    expected = [Affiliation(id=60104842, name='College of Engineering',
                    city='Pittsburgh', country='United States')]
    assert ab1.affiliation == expected
    assert ab3.affiliation is None
    assert ab8.affiliation is None


def test_aggregationType():
    assert ab1.aggregationType == 'Journal'
    assert ab2.aggregationType == 'Conference Proceeding'
    assert ab8.aggregationType is None


def test_authkeywords():
    assert ab1.authkeywords is None
    expected = ['Bayesian analysis', 'Seasonality',
                'Structural breaks', 'Unit roots']
    assert ab4.authkeywords == expected
    assert ab8.authkeywords is None


def test_authorgroup():
    # Test FULL document w.o. collaboration
    expected = [AuthorGroup(affiliation_id=60104842, collaboration_id=None, dptid=None,
        organization='Department of Chemical Engineering, Carnegie Mellon University',
        city='Pittsburgh', postalcode='15213', addresspart='5000 Forbes Avenue',
        country='United States', auid=7004212771, orcid=None, indexed_name='Kitchin J.R.',
        surname='Kitchin', given_name='John R.')]
    assert ab1.authorgroup == expected
    # Test FULL document w. 1 collaboration
    expected = AuthorGroup(affiliation_id=None, collaboration_id=None, dptid=None, organization=None, 
        city=None, postalcode=None, addresspart=None, country=None, auid=7403019450, 
        orcid=None, indexed_name='Ahn J.K.', surname='Ahn', given_name='J.K.')
    assert ab9.authorgroup[0] == expected
    # Test FULL document w. list of collaborations
    expected = AuthorGroup(affiliation_id=None, collaboration_id='S0920379618302370-8f4b482a50491834a3f938b012344bfd',
        dptid=None, organization=None, city=None, postalcode=None, addresspart=None, country=None,
        auid=None, orcid=None, indexed_name='JET Contributors', surname=None, given_name=None)
    assert ab12.authorgroup[-1] == expected
    # Test REF view
    assert ab8.authorgroup is None


def test_authors():
    expected = [Author(auid=7004212771, indexed_name='Kitchin J.R.',
                surname='Kitchin', given_name='John R.',
                affiliation='60104842')]
    assert ab1.authors == expected
    assert ab8.authors is None
    assert ar10.authors is None


def test_citedby_count():
    expected = 5
    assert ab1.citedby_count >= expected
    assert ab8.citedby_count is None


def test_citedby_link():
    expected = 'https://www.scopus.com/inward/citedby.uri?partnerID=HzOxMe3b'\
        '&scp=84930616647&origin=inward'
    assert ab1.citedby_link == expected
    assert ab8.citedby_link is None


def test_chemials():
    received = ab6.chemicals
    assert isinstance(received, list)
    assert len(received) == 6
    expected = Chemical(source='esbd', cas_registry_number='126547-89-5',
        chemical_name='intercellular adhesion molecule 1')
    assert expected in received
    assert ab3.chemicals is None
    assert ab8.chemicals is None


def test_confcode():
    assert ab2.confcode == 44367
    assert ab8.confcode is None


def test_confdate():
    assert ab2.confdate == ((1995, 12, 13), (1995, 12, 15))
    assert ab8.confdate is None


def test_conflocation():
    assert ab2.conflocation == 'New Orleans, LA, USA'
    assert ab8.conflocation is None


def test_confname():
    expected2 = "Proceedings of the 1995 34th IEEE Conference on Decision "\
                "and Control. Part 1 (of 4)"
    assert ab2.confname == expected2
    assert ab3.confname is None
    expected7 = '20th Symposium on Design, Test, Integration and Packaging '\
                'of MEMS and MOEMS, DTIP 2018'
    assert ab7.confname == expected7
    assert ab8.confname is None


def test_confsponsor():
    assert ab2.confsponsor == 'IEEE'
    assert ab3.confsponsor is None
    expected7 = ['ARTOV.IMM.CNR.IT', 'CMP.IMAG.FR', 'CNRS.FR',
                 'EPS.IEEE.ORG', 'LIRMM.FR']
    assert ab7.confsponsor == expected7
    assert ab8.confsponsor is None


def test_contributor_group():
    expected = Contributor(given_name='Romolo', initials='R.', surname='Marcelli',
                    indexed_name='Marcelli R.', role='edit')
    received = ab7.contributor_group
    assert len(received) == 7
    assert expected in received
    assert ab3.contributor_group is None
    assert ab8.contributor_group is None

    expected_13 = Contributor(given_name='Constantinos', initials='C.', surname='Antoniou',
                    indexed_name='Antoniou C.', role='edit')
    assert len(ab13.contributor_group) == 4
    assert ab13.contributor_group[0] == expected_13


def test_copyright():
    assert ab8.copyright is None
    expected = "Copyright 2021 Elsevier B.V., All rights reserved."
    assert ab9.copyright == expected


def test_copyright_type():
    assert ab8.copyright_type is None
    assert ab9.copyright_type == "Elsevier"


def test_correspondence():
    expected2 = Correspondence(surname='Boukas', initials='E.K.',
        organization='Ecole Polytechnique de Montreal', country='Canada',
        city_group='Montreal')
    assert ab2.correspondence[0] == expected2
    assert ab3.correspondence is None
    assert ab8.correspondence is None


def test_coverDate():
    assert ab1.coverDate == '2015-06-05'
    assert ab8.coverDate is None


def test_date_created():
    assert ab8.date_created is None
    assert ab9.date_created == (2021, 9, 14)


def test_description():
    expected = 'In this paper we propose a Bayesian analysis of seasonal '\
               'unit roots in quarterly observed time series. Seasonal unit '\
               'root processes are useful to describe economic series with '\
               'changing seasonal fluctuations. A natural alternative model '\
               'for similar purposes contains deterministic seasonal mean '\
               'shifts instead of seasonal stochastic trends. This leads to '\
               'analysing seasonal unit roots in the presence of mean '\
               'shifts using Bayesian techniques. Our method is illustrated '\
               'using several simulated and empirical data.'
    assert ab4.description == expected
    assert ab8.description is None


def test_doi():
    assert ab1.doi == '10.1021/acscatal.5b00538'
    assert ab4.doi == '10.1016/s0304-4076(97)00018-3'
    assert ab8.doi is None


def test_eid():
    assert ab1.eid == '2-s2.0-84930616647'
    assert ab8.eid is None


def test_endingPage():
    assert ab1.endingPage == '3899'
    assert ab8.endingPage is None


def test_entitlement():
    assert ab8.document_entitlement_status is None
    assert ab9.document_entitlement_status is None
    assert ar10.document_entitlement_status == 'ENTITLED'


def test_funding():
    received = ab6.funding
    assert isinstance(received, list)
    assert len(received) == 8
    expected6 = Funding(
        agency='Deutsche Forschungsgemeinschaft',
        agency_id='http://data.elsevier.com/vocabulary/SciValFunders/501100001659',
        string='German Research Foundation', acronym='DFG',
        funding_id=['SFB685'],
        country='http://sws.geonames.org/2921044/')
    assert expected6 in received
    assert ab5.funding is None
    assert ab8.funding is None


def test_funding_text():
    e = 'ACKNOWLEDGMENTS. We thank Dieter Blaas for providing ICAM-1–specific '\
        'antiserum (supersup). This work was supported by Netherlands '\
        'Organization for Scientific Research Grant NWO-VICI-91812628 '\
        '(to F.J.M.v.K.), by German Research Foundation Grant SFB685 '\
        '(to T.S. and G.Z.), and Wellcome Trust PhD Studentship support '\
        '102572/B/13/Z (to D.L.H.). All EM was performed in the Astbury '\
        'Biostructure Laboratory, which was funded by the University of '\
        'Leeds and the Wellcome Trust (108466/Z/15/Z).'
    assert ab6.funding_text == e
    assert ab8.funding_text is None


def test_get_bibtex():
    e = '@article{Kaufmann1991FairnessPricing,\n  author = {Patrick J. '\
        'Kaufmann and Gwen Ortmeyer and N. Craig Smith},\n  title = {{'\
        'Fairness in consumer pricing}},\n  journal = {Journal of Consumer '\
        'Policy},\n  year = {1991},\n  volume = {14},\n  number = {2},\n  '\
        'pages = {117-140},\n  doi = {10.1007/BF00381915}}'
    assert ab3.get_bibtex() == e


def test_get_latex():
    e = 'Philip Hans Franses, Henk Hoekh and Richard Paap, \\textit{'\
        'Bayesian analysis of seasonal unit roots and seasonal mean shifts}, '\
        'Journal of Econometrics, \\textbf{78(2)}, pp. 359-380 (1997). \\'\
        'href{https://doi.org/10.1016/s0304-4076(97)00018-3}{doi:10.1016/'\
        's0304-4076(97)00018-3}, \\href{https://www.scopus.com/inward/'\
        'record.uri?partnerID=HzOxMe3b&scp=0000016206&origin=inward}'\
        '{scopus:2-s2.0-0000016206}.'
    assert ab4.get_latex() == e


def test_get_ris():
    e = (
        "TY  - JOUR\nTI  - Examples of effective data sharing in scientific "
        "publishing\nJO  - ACS Catalysis\nDA  - 2015-06-05\nPY  - "
        "2015\nAU  - Kitchin J.R.\nDO  - 10.1021/acscatal.5b00538\n"
        "UR  - https://doi.org/10.1021/acscatal.5b00538\nVL  - 5\n"
        "IS  - 6\nSP  - 3894-3899\nER  - \n\n"
    )
    assert ab1.get_ris() == e


def test_get_html():
    e = '<a href="https://www.scopus.com/authid/detail.url?origin=Author'\
        'Profile&authorId=7201922462">Patrick J. Kaufmann</a>, <a href="'\
        'https://www.scopus.com/authid/detail.url?origin=AuthorProfile&'\
        'authorId=16430389100">Gwen Ortmeyer</a> and <a href="https://'\
        'www.scopus.com/authid/detail.url?origin=AuthorProfile&authorId='\
        '57225963896">N. Craig Smith</a>, <a href="https://www.scopus.com/'\
        'inward/record.uri?partnerID=HzOxMe3b&scp=0001270077&origin=inward">'\
        'Fairness in consumer pricing</a>, <a href="https://www.scopus.com/'\
        'source/sourceInfo.url?sourceId=130073">Journal of Consumer Policy'\
        '</a>, <b>14(2)</b>, pp. 117-140, (1991). <a href="https://doi.org/'\
        '10.1007/BF00381915">doi:10.1007/BF00381915</a>.'
    assert ab3.get_html() == e


def test_isbn():
    assert ab3.isbn is None
    assert ab5.isbn == ('0203881486', '9780203881484')
    assert ab8.isbn is None


def test_issn():
    expected1 = ISSN(print='21555435', electronic=None)
    assert ab1.issn == expected1
    expected3 = ISSN(print='03425843', electronic='15730700')
    assert ab3.issn == expected3
    assert ab5.issn is None
    assert ab8.issn is None


def test_identifier():
    assert ab1.identifier == 84930616647
    assert ab8.identifier is None


def test_idxterms():
    expected = ['Control variables', 'Cost function',
                'Hamilton-Jacobi-Isaacs equation', 'Machine capacity',
                'Stochastic manufacturing systems', 'Value function']
    assert ab2.idxterms == expected
    assert ab4.idxterms is None


def test_issueIdentifier():
    assert ab1.issueIdentifier == '6'
    assert ab8.issueIdentifier is None


def test_issuetitle():
    assert ab3.issuetitle == 'Law, Economics and Behavioural Sciences'
    assert ab8.issuetitle is None


def test_language():
    assert ab1.language == 'eng'
    assert ab8.language is None


def test_openaccess():
    assert ab5.openaccess == 0
    assert ab6.openaccess == 1
    assert ab7.openaccess == 0
    assert ab8.openaccess is None


def test_openaccessFlag():
    assert ab5.openaccessFlag == False
    assert ab6.openaccessFlag == True
    assert ab7.openaccessFlag == False
    assert ab8.openaccessFlag is None


def test_pageRange():
    assert ab1.pageRange == '3894-3899'
    assert ab8.pageRange is None


def test_pii():
    assert ab4.pii == 'S0304407697000183'
    assert ab5.pii is None
    assert ab6.pii is None


def test_publicationName():
    assert ab1.publicationName == 'ACS Catalysis'
    assert ab8.publicationName is None


def test_publisher():
    assert ab1.publisher == 'American Chemical Society'
    assert ab8.publisher is None


def test_publisheraddress():
    assert ab2.publisheraddress == 'Piscataway, NJ, United States'
    assert ab8.publisheraddress is None


def test_pubmed_id():
    assert ab6.pubmed_id == 29284752
    assert ab7.pubmed_id is None


def test_refcount():
    assert ab4.refcount == 18
    assert ab8.refcount == 48
    assert ab11.refcount is None


def test_references_full():
    fulltext1 = 'Implementing Reproducible Research; Stodden, V.; Leisch, '\
                'F.; Peng, R. D., Eds., Chapman and Hall/CRC: London, 2014.'
    expected1 = Reference(position='22', id='85055586929', doi=None, title=None,
        authors='Stodden, V.; Leisch, F.; Peng, R.D.', authors_auid=None,
        authors_affiliationid=None, fulltext=fulltext1, coverDate=None,
        sourcetitle='Implementing Reproducible Research', type=None,
        publicationyear='2014', volume=None, issue=None, first=None,
        last=None, citedbycount=None, text='Eds. Chapman and Hall/CRC: London.')
    expected7 = Reference(position='1', id='85050215448', doi=None, title=None,
        authors='', authors_auid=None, authors_affiliationid=None,
        sourcetitle=None, publicationyear=None, coverDate=None, volume=None,
        issue=None, first=None, last=None, citedbycount=None, type=None,
        text='accessed 27 June 2017',
        fulltext='www. hexoskin. com, accessed 27 June 2017')
    assert ab1.references[-1] == expected1
    assert ab2.references is None
    assert ab7.references[0] == expected7


def test_references_ref():
    assert len(ab8.references) == 48
    expected8 = Reference(position='47', id='77953310709',
        doi='10.1109/INFCOM.2010.5462174', text=None, fulltext=None,
        title='Achieving secure, scalable, and fine-grained data access control in cloud computing',
        authors='Yu, Shucheng; Lou, Wenjing; Wang, Cong; Ren, Kui',
        authors_auid='55636591800; 7006030576; 35106222100; 8396435500',
        authors_affiliationid='60157577; 60157577; 60002873; 60002873',
        sourcetitle='Proceedings - IEEE INFOCOM',
        publicationyear=None, coverDate='2010-06-15', volume=None, issue=None,
        first=None, last=None, citedbycount='0', type='resolvedReference')
    assert ab8.references[-2]._replace(citedbycount="0") == expected8
    assert int(ab8.references[-2].citedbycount) >= 0
    assert ab11.references is None


def test_scopus_link():
    expected = 'https://www.scopus.com/inward/record.uri?partnerID=HzOxMe3b&'\
        'scp=84930616647&origin=inward'
    assert ab1.scopus_link == expected


def test_self_link():
    expected = 'https://api.elsevier.com/content/abstract/scopus_id/84930616647'
    assert ab1.self_link == expected
    assert ab8.self_link is None


def test_sequencebank():
    received = ab6.sequencebank
    assert isinstance(received, list)
    assert len(received) == 3
    expected = Sequencebank(name='GENBANK', sequence_number='MG272373',
                    type='referenced')
    assert expected in received
    assert ab3.sequencebank is None
    assert ab8.sequencebank is None


def test_source_id():
    assert ab1.source_id == 19700188320
    assert ab8.source_id is None


def test_sourcetitle_abbreviation():
    assert ab1.sourcetitle_abbreviation == 'ACS Catal.'
    assert ab8.sourcetitle_abbreviation is None


def test_srctype():
    assert ab1.srctype == 'j'
    assert ab2.srctype == 'p'
    assert ab8.srctype is None


def test_startingPage():
    assert ab1.startingPage == '3894'
    assert ab8.startingPage is None


def test_subject_areas():
    expected = [Area(area='Catalysis', abbreviation='CENG', code=1503),
                Area(area='Chemistry (all)', abbreviation='CHEM', code=1600)]
    assert ab1.subject_areas == expected
    assert ab8.subject_areas is None
    expected = [Area(area='Nuclear and High Energy Physics',
                     abbreviation='PHYS', code=3106)]
    assert ab9.subject_areas == expected


def test_subtype():
    assert ab1.subtype == "re"
    assert ab2.subtype == "cp"
    assert ab5.subtype == "bk"
    assert ab6.subtype == "ar"
    assert ar10.subtype is None


def test_subtypedescription():
    assert ab1.subtypedescription == "Review"
    assert ab2.subtypedescription == "Conference Paper"
    assert ab5.subtypedescription == "Book"
    assert ab6.subtypedescription == "Article"
    assert ar10.subtypedescription is None


def test_title():
    expected = 'Examples of effective data sharing in scientific publishing'
    assert ab1.title == expected
    assert ab8.title is None
    assert ar10.title is None


def test_url():
    expected = 'https://api.elsevier.com/content/abstract/scopus_id/84930616647'
    assert ab1.url == expected
    assert ab8.url is None


def test_volume():
    assert ab1.volume == '5'
    assert ab8.volume is None


def test_website():
    assert ab1.website == 'http://pubs.acs.org/page/accacs/about.html'
    assert ab2.website is None
    assert ab8.website is None



"""Tests for sciencedirect.ArticleEntitlement"""

from pybliometrics.sciencedirect import ArticleEntitlement, init

init()

ae_1 = ArticleEntitlement('10.1016/j.reseneeco.2015.06.001', view='FULL', id_type='doi', refresh=30)
ae_2 = ArticleEntitlement('S0140988320302814', view='FULL', refresh=30)


def test_all_fields():
    """Test all propoerties of ArticleEntitlement."""
    assert ae_1.doi == '10.1016/j.reseneeco.2015.06.001'
    assert ae_1.eid == '1-s2.0-S092876551500038X'
    assert ae_1.entitled is True
    assert ae_1.identifier == 'http://dx.doi.org/10.1016/j.reseneeco.2015.06.001'
    assert ae_1.link == 'https://www.sciencedirect.com/science/article/pii/S092876551500038X'
    assert ae_1.message == 'Requestor is entitled to the requested resource'
    assert ae_1.pii == 'S0928-7655(15)00038-X'
    assert ae_1.pii_norm == 'S092876551500038X'
    assert ae_1.pubmed_id is None
    assert ae_1.scopus_id == '84935028440'
    assert ae_1.status == 'found'
    assert ae_1.url == 'https://api.elsevier.com/content/article/pii/S092876551500038X'


def test_str():
    """Test print message."""
    expected_str = 'Requestor is entitled to the requested resource with doi: 10.1016/j.eneco.2020.104941'
    assert ae_2.__str__() == expected_str



"""Tests for `scopus.CitationOverview` module."""

from pybliometrics.scopus import init
from pybliometrics.scopus.abstract_citation import Author, CitationOverview

init()


co_eid = CitationOverview(["84930616647", "85068268027"],
                          refresh=30, date="2016-2020")
co_doi = CitationOverview(["10.1016/j.softx.2019.100263"],
                          id_type="doi", refresh=30, date="2016-2020")

def test_authors():
    url = 'https://api.elsevier.com/content/author/author_id/7004212771'
    john = Author(name='J.R., Kitchin, John R.', surname='Kitchin',
                  initials='J.R.', id='7004212771', url=url)
    assert co_eid.authors[0][1] == john
    assert co_eid.authors[1] == [john]
    assert co_doi.authors[0][1] == john


def test_cc():
    expected0 = [(2016, 0), (2017, 0), (2018, 0), (2019, 0), (2020, 6)]
    expected1 = [(2016, 4), (2017, 2), (2018, 3), (2019, 2), (2020, 2)]
    assert co_eid.cc == [expected0, expected1]
    assert co_doi.cc == [expected0]


def test_citationType_long():
    assert co_eid.citationType_long == ['ARTICLE', 'REVIEW']
    assert co_doi.citationType_long == ['ARTICLE']


def test_citationType_short():
    assert co_eid.citationType_short == ['ARTICLE', 'REVIEW']
    assert co_doi.citationType_short == ['ARTICLE']


def test_columnTotal():
    assert co_eid.columnTotal == [4, 2, 3, 2, 8]
    assert co_doi.columnTotal == [0, 0, 0, 0, 6]


def test_doi():
    expected = ['10.1016/j.softx.2019.100263', '10.1021/acscatal.5b00538']
    assert co_eid.doi == expected
    assert co_doi.doi == [expected[0]]


def test_endingPage():
    assert co_eid.endingPage == [None, '3899']
    assert co_doi.endingPage is None


def test_grandTotal():
    assert co_eid.grandTotal >= 29
    assert co_doi.grandTotal >= 16


def test_h_index():
    assert co_eid.h_index == 2
    assert co_doi.h_index == 1


def test_issn():
    expected = ['2352-7110', '2155-5435']
    assert co_eid.issn == expected
    assert co_doi.issn == [expected[0]]


def test_issueIdentifier():
    assert co_eid.issueIdentifier == [None, '6']
    assert co_doi.issueIdentifier is None


def test_laterColumnTotal():
    assert co_eid.laterColumnTotal >= 18
    assert co_doi.laterColumnTotal >= 16


def test_lcc():
    assert co_eid.lcc[0] >= 1
    assert co_eid.lcc[1] >= 1
    assert co_doi.lcc[0] >= 1


def test_pcc():
    assert co_eid.pcc == [0, 0]
    assert co_doi.pcc == [0]


def test_pii():
    expected = ['S2352711019300573', None]
    assert co_eid.pii == expected
    assert co_doi.pii == [expected[0]]


def test_prevColumnTotal():
    assert co_eid.prevColumnTotal == 0
    assert co_doi.prevColumnTotal == 0


def test_publicationName():
    expected = ['SoftwareX', 'ACS Catalysis']
    assert co_eid.publicationName == expected
    assert co_doi.publicationName == [expected[0]]


def test_rangeColumnTotal():
    assert co_eid.rangeColumnTotal == 19
    assert co_doi.rangeColumnTotal == 6


def test_rangeCount():
    assert co_eid.rangeCount[0] >= 6
    assert co_eid.rangeCount[1] >= 6
    assert co_doi.rangeCount[0] >= 6


def test_rowTotal():
    assert co_eid.rowTotal[0] >= 10
    assert co_eid.rowTotal[1] >= 10
    assert co_doi.rowTotal[0] >= 10


def test_scopus_id():
    expected = [85068268027, 84930616647]
    assert co_eid.scopus_id == expected
    assert co_doi.scopus_id == [expected[0]]


def test_startingPage():
    assert co_eid.startingPage == [None, '3894']
    assert co_doi.startingPage == None


def test_title():
    expected = ['pybliometrics: Scriptable bibliometrics using a Python interface to Scopus',
                'Examples of effective data sharing in scientific publishing']
    assert co_eid.title == expected
    assert co_doi.title == [expected[0]]


def test_url():
    expected = ['https://api.elsevier.com/content/abstract/scopus_id/85068268027',
                'https://api.elsevier.com/content/abstract/scopus_id/84930616647']
    assert co_eid.url == expected
    assert co_doi.url == [expected[0]]


def test_volume():
    assert co_eid.volume == ['10', '5']
    assert co_doi.volume == ['10']



import warnings
from configparser import ConfigParser, NoOptionError, NoSectionError
from collections import deque
from pathlib import Path

from pybliometrics.utils.constants import CONFIG_FILE, RATELIMITS, DEFAULT_PATHS, VIEWS
from pybliometrics.utils.create_config import create_config

CONFIG = None
CUSTOM_KEYS = None
CUSTOM_INSTTOKENS = None

_throttling_params = {k: deque(maxlen=v) for k, v in RATELIMITS.items()}


def init(config_path: str | Path | None = None,
         keys: list[str] | None = None,
         inst_tokens: list[str] | None = None,
         config_dir: str | Path | None = None) -> None:
    """
    Function to initialize the pybliometrics library. For more information refer to the
    `official documentation <https://pybliometrics.readthedocs.io/en/stable/configuration.html>`_.
    
    :param config_path: Path to the configuration file.  If None, defaults to
                        pybliometrics.utils.constants.CONFIG_FILE.
    :param keys: List of API keys.
    :param inst_tokens: List of corresponding InstTokens. The order must match that
                        of `keys` to avoid errors.
    :param config_dir: (Deprecated) Path to the configuration file. Use `config_path` instead.

    :raises NoSectionError: If the required sections (Directories, Authentication, Request)
                            do not exist.
    :raises ValueError: If there are no or fewer API keys than InstTokens present.
    """
    global CONFIG
    global CUSTOM_KEYS
    global CUSTOM_INSTTOKENS

    # Deprecation inserted in 4.2
    if config_dir is not None:
        msg = "The 'config_dir' parameter is deprecated and will be removed in " \
              "a future version. Please use 'config_path' instead."
        warnings.warn(msg, DeprecationWarning, stacklevel=2)
        if config_path is None:
            config_path = config_dir

    if not config_path:
        config_path = CONFIG_FILE
    config_path = Path(config_path)

    if not config_path.exists():
        CONFIG = create_config(config_path, keys, inst_tokens)
    else:
        CONFIG = ConfigParser()
        CONFIG.optionxform = str
        CONFIG.read(config_path)

    check_sections(CONFIG)
    check_default_paths(CONFIG, config_path)
    create_cache_folders(CONFIG)

    CUSTOM_KEYS = keys
    CUSTOM_INSTTOKENS = inst_tokens
    check_keys_tokens()


def check_sections(config: ConfigParser) -> None:
    """Auxiliary function to check if all sections exist."""
    for section in ['Directories', 'Authentication', 'Requests']:
        if not config.has_section(section):
            raise NoSectionError(section)


def check_default_paths(config: ConfigParser, config_path: Path) -> None:
    """Auxiliary function to check if default cache paths exist.
    If not, the paths are writen in the config.
    """
    for api, path in DEFAULT_PATHS.items():
        if not config.has_option('Directories', api):
            config.set('Directories', api, str(path))
            with open(config_path, 'w', encoding='utf-8') as ouf:
                config.write(ouf)


def check_keys_tokens() -> None:
    """Auxiliary function to check if API keys or InstTokens are set."""
    keys = get_keys()
    insttokens = get_insttokens()
    # 3 problematic cases
    no_keys_no_insttokens = not keys and not insttokens
    insttokens_no_keys = insttokens and not keys
    keys_and_insttokens = keys and insttokens
    keys_tokens_diff = len(keys) - len(insttokens)

    if no_keys_no_insttokens:
        raise ValueError('No API keys or InstTokens found. '
                         'Please provide at least one API key or InstToken. '
                         'For more information visit: '
                         'https://pybliometrics.readthedocs.io/en/stable/configuration.html')
    elif insttokens_no_keys:
        raise ValueError('InstTokens found but not corresponding API keys. '
                         'Please provide the API keys that correspond to the InstTokens. '
                         'For more information visit: '
                         'https://pybliometrics.readthedocs.io/en/stable/configuration.html')
    elif keys_and_insttokens:
        if keys_tokens_diff < 0:
            raise ValueError('More InstTokens than API keys found. '
                             'Please provide all the API keys that correspond to the InstTokens. '
                             'For more information visit: '
                             'https://pybliometrics.readthedocs.io/en/stable/configuration.html')


def create_cache_folders(config: ConfigParser) -> None:
    """Auxiliary function to create cache folders."""
    for api, path in config.items('Directories'):
        views = VIEWS.get(api, [""])  # empty string
        for view in views:
            view_path = Path(path, view)
            view_path.mkdir(parents=True, exist_ok=True)


def get_config() -> ConfigParser:
    """Function to get the config parser."""
    if not CONFIG:
        raise FileNotFoundError('No configuration file found.'
                                'Please initialize Pybliometrics with init().\n'
                                'For more information visit: '
                                'https://pybliometrics.readthedocs.io/en/stable/configuration.html')
    return CONFIG


def get_insttokens() -> list[tuple[str, str]]:
    """Function to get the InstToken and overwrite InstToken in config if needed."""
    inst_tokens = []
    if CUSTOM_INSTTOKENS:
        inst_tokens = CUSTOM_INSTTOKENS
    else:
        if not CUSTOM_KEYS: # if custom keys are set, config inst tokens are not needed
            try:
                raw_token_text = CONFIG.get('Authentication', 'InstToken')
                inst_tokens = [k.strip() for k in raw_token_text.split(",")]
            except NoOptionError:
                inst_tokens = []

    #key_token_pairs = list(zip(get_all_keys(), inst_tokens))
    return inst_tokens


def get_keys() -> list[str]:
    """Function to get all the API keys and overwrite keys in config if needed."""
    if CUSTOM_KEYS:
        keys = CUSTOM_KEYS
    else:
        try:
            raw_keys_text = CONFIG.get('Authentication', 'APIKey')
            keys = [k.strip() for k in raw_keys_text.split(",")]
        except NoOptionError:
            keys = []
    return keys



"""Tests for `scopus.AuthorSearch` module."""

from pybliometrics.scopus import AuthorSearch, init
from pybliometrics.scopus.author_search import Author

init()

s1 = AuthorSearch('authlast(selten) and authfirst(reinhard)', refresh=30)
s2 = AuthorSearch('authlast(selten)', download=False, refresh=True)


def test_authors():
    assert isinstance(s1.authors, list)
    assert len(s1.authors) >= 1
    expected = Author(eid='9-s2.0-6602907525', orcid=None, surname='Selten',
        initials='R.', givenname='Reinhard', affiliation='Universität Bonn',
        documents=76, affiliation_id='60007493', city='Bonn',
        country='Germany', areas='ECON (72); BUSI (8)')
    assert s1.authors[0] == expected


def test_authors_nodownload():
    # Only works if query hasn't been cached
    assert s2.authors is None


def test_results_size():
    received1 = s1.get_results_size()
    assert received1 >= 1
    received2 = s2.get_results_size()
    assert received2 >= 25



"""Tests for `scopus.PlumXMetrics` module."""

from pybliometrics.scopus import PlumXMetrics, init

init()

# A published paper. All PlumX Metrics categories are present.
m1 = PlumXMetrics('2-s2.0-85054706190', 'elsevierId', refresh=30)
# A published paper. No Usage and Social Media metrics.
m2 = PlumXMetrics('10.2307/2281868', 'doi', refresh=30)
# A published paper. No PlumX Metrics.
m3 = PlumXMetrics('2-s2.0-84950369844', 'elsevierId', refresh=30)
# A published book. No Social Media metrics.
m4 = PlumXMetrics('9783540783893', 'isbn', refresh=30)
# A SSRN paper. No mention metrics.
m5 = PlumXMetrics('3320470', 'ssrnId', refresh=30)


def test_category_totals():
    m1_received = sorted([c.name  for c in m1.category_totals])
    m1_expected = ['capture', 'citation', 'mention', 'socialMedia']
    assert m1_received == m1_expected
    m2_received = sorted([c.name  for c in m2.category_totals])
    m2_expected = ['capture', 'citation', 'mention']
    assert m2_received == m2_expected
    m3_received = sorted([c.name  for c in m3.category_totals])
    assert m3_received == ['citation']
    m5_received = sorted([c.name  for c in m5.category_totals])
    m5_expected = ['capture', 'citation', 'socialMedia', 'usage']
    assert m5_received == m5_expected
    for plumx in (m1, m2, m4, m5):
        m_fields = set(field for ntup in plumx.category_totals for
                       field in ntup._fields)
        assert m_fields, {'name', 'total'}
        zero_totals = [i.total for i in plumx.category_totals if i.total <= 0]
        assert zero_totals == []


def test_capture():
    assert m3.capture is None
    expected = {'name', 'total'}
    for plumx in (m1, m2, m4, m5):
        assert isinstance(plumx.capture, list)
        assert len(plumx.capture) > 0
        m_fields = set(field for ntup in plumx.capture for field in ntup._fields)
        assert m_fields == expected
        zero_totals = [i.total for i in plumx.capture if i.total <= 0] 
        assert zero_totals == []


def test_citation():
    assert m3.citation[0].total == 0
    expected = {'name', 'total'}
    for plumx in (m1, m2, m4, m5):
        assert isinstance(plumx.citation, list)
        assert len(plumx.citation) > 0
        m_fields = set(field for ntup in plumx.citation for field in ntup._fields)
        assert m_fields == expected
        zero_totals = [i.total for i in plumx.citation if i.total <= 0]
        assert zero_totals == []


def test_mention():
    assert m3.mention is None
    assert m5.mention is None
    expected = {'name', 'total'}
    for plumx in (m1, m2, m4):
        print(plumx)
        assert isinstance(plumx.mention, list)
        assert len(plumx.mention) > 0
        m_fields = set(field for ntup in plumx.mention for field in ntup._fields)
        assert m_fields == expected
        zero_totals = [i.total for i in plumx.mention if i.total <= 0]
        assert zero_totals == []


def test_social_media():
    assert m2.social_media is None
    assert m3.social_media is None
    assert m4.social_media is None
    expected = {'name', 'total'}
    for plumx in (m1, m5):
        assert isinstance(plumx.social_media, list)
        assert len(plumx.social_media) > 0
        m_fields = set(field for ntup in plumx.social_media for field in ntup._fields)
        assert m_fields == expected
        zero_totals = [i.total for i in plumx.social_media if i.total <= 0]
        assert zero_totals == []


def test_usage():
    assert m2.usage is None
    assert m3.usage is None
    assert m4.usage is None
    assert isinstance(m5.usage, list)
    assert len(m5.usage) > 0
    m_fields = set(field for ntup in m5.usage for field in ntup._fields)
    assert m_fields == {'name', 'total'}
    zero_totals = [i.total for i in m5.usage if i.total <= 0]
    assert zero_totals == []



"""Tests for `scopus.AuthorRetrieval` module."""

from pybliometrics.scopus import AuthorRetrieval, init
from pybliometrics.scopus.author_retrieval import Affiliation, Coauthor

init()

metrics = AuthorRetrieval("7004212771", refresh=30, view="METRICS")
light = AuthorRetrieval("7004212771", refresh=30, view="LIGHT")
standard = AuthorRetrieval("7004212771", refresh=30, view="STANDARD")
enhanced = AuthorRetrieval("7004212771", refresh=30, view="ENHANCED")
entitled = AuthorRetrieval(36009348900, view='ENTITLED')


def test_affiliation_current():
    assert metrics.affiliation_current is None

    expected_std_enh = Affiliation(id=60027950, parent=None, type='parent',
        relationship='author', afdispname=None, preferred_name='Carnegie Mellon University',
        parent_preferred_name=None, country_code='usa', country='United States',
        address_part='5000 Forbes Avenue', city='Pittsburgh', state='PA',
        postal_code='15213-3890', org_domain='cmu.edu', org_URL='https://www.cmu.edu/')

    expected_lgh = Affiliation(id=None, parent=None, type=None,
        relationship=None, afdispname=None, preferred_name='Carnegie Mellon University',
        parent_preferred_name=None, country_code=None, country='United States',
        address_part=None, city='Pittsburgh', state=None,
        postal_code=None, org_domain=None, org_URL=None)

    for a in (standard, enhanced, light):
        received = a.affiliation_current
        assert isinstance(received, list)
        assert len(received) >= 1

    for a in (standard, enhanced):
        received = a.affiliation_current
        assert expected_std_enh in received

    assert expected_lgh in light.affiliation_current


def test_affiliation_history():
    assert metrics.affiliation_history is None
    assert light.affiliation_history is None
    expected = Affiliation(id=60008644, parent=None, type='parent',
        relationship='author', afdispname=None,
        preferred_name='Fritz Haber Institute of the Max Planck Society',
        parent_preferred_name=None, country_code='deu', country='Germany',
        address_part='Faradayweg 4-6', city='Berlin', state=None,
        postal_code='14195', org_domain='fhi.mpg.de',
        org_URL='https://www.fhi.mpg.de/')
    for a in (standard, enhanced):
        received = a.affiliation_history
        assert isinstance(received, list)
        assert len(received) >= 10
        assert expected in received


def test_alias():
    assert metrics.alias is None
    assert light.alias is None
    assert standard.alias is None
    assert enhanced.alias is None


def test_citation_count():
    expected = 13600
    assert metrics.citation_count >= expected
    assert light.citation_count >= expected
    assert standard.citation_count >= expected
    assert enhanced.citation_count >= expected


def test_cited_by_count():
    expected = 10900
    assert metrics.cited_by_count >= expected
    assert light.cited_by_count >= expected
    assert standard.cited_by_count >= expected
    assert enhanced.cited_by_count >= expected


def test_classificationgroup():
    assert metrics.classificationgroup is None
    assert light.classificationgroup is None
    for a in (standard, enhanced):
        received = a.classificationgroup
        assert isinstance(received, list)
        assert len(received) > 0
        assert (1906, 1) in received  # frequency might differ


def test_coauthor_count():
    assert light.coauthor_count is None
    assert standard.coauthor_count is None
    expected = 175
    assert metrics.coauthor_count >= expected
    assert enhanced.coauthor_count >= expected


def test_coauthor_link():
    assert metrics.coauthor_link is None
    assert light.coauthor_link is None
    assert standard.coauthor_link is None
    expected = 'http://api.elsevier.com/content/search/author?co-author=7004212771'
    assert enhanced.coauthor_link == expected


def test_date_created():
    assert metrics.date_created is None
    assert light.date_created is None
    assert standard.date_created, (2005, 12, 3)
    assert enhanced.date_created, (2005, 12, 3)


def test_document_count():
    expected = 106
    assert light.document_count >= expected
    assert standard.document_count >= expected
    assert metrics.document_count >= expected
    assert enhanced.document_count >= expected


def test_eid():
    assert metrics.eid is None
    assert light.eid, '9-s2.0-7004212771'
    assert standard.eid, '9-s2.0-7004212771'
    assert enhanced.eid, '9-s2.0-7004212771'


def test_estimate_uniqueness():
    assert metrics.estimate_uniqueness() == 2
    assert light.estimate_uniqueness() == 2
    assert standard.estimate_uniqueness() == 1
    assert enhanced.estimate_uniqueness() == 1


def test_entitlement():
    assert metrics.document_entitlement_status is None
    assert light.document_entitlement_status is None
    assert standard.document_entitlement_status is None
    assert enhanced.document_entitlement_status is None
    assert entitled.document_entitlement_status == 'ENTITLED'


def test_given_name():
    assert metrics.given_name is None
    assert light.given_name is None
    assert standard.given_name == 'John R.'
    assert enhanced.given_name == 'John R.'


def get_coauthors():
    assert metrics.get_coauthors() is None
    assert light.get_coauthors() is None
    assert standard.get_coauthors() is None
    received = enhanced.get_coauthors()
    assert isinstance(received, list)
    assert len(received) > 155
    expected = Coauthor(surname='Rose', given_name='Michael E.', id=57209617104,
        areas='Computer Science (all)', affiliation_id='60105007',
        name='Max-Planck-Institut für Innovation und Wettbewerb',
        city='Munich', country='Germany')
    assert expected in received


def test_get_documents():
    subtypes = {'re', 'ed', 'no'}
    received = enhanced.get_documents(subtypes)
    assert len(received) == 8


def test_get_document_eids():
    expected = 100
    assert len(enhanced.get_document_eids()) >= expected


def test_h_index():
    assert light.h_index is None
    assert standard.h_index is None
    expected = 34
    assert metrics.h_index >= expected
    assert enhanced.h_index >= expected
    assert entitled.h_index is None


def test_historical_identifier():
    assert metrics.historical_identifier is None
    assert light.historical_identifier is None
    assert standard.historical_identifier is None
    assert enhanced.historical_identifier == None


def test_identifier():
    expected = 7004212771
    assert metrics.identifier == expected
    assert light.identifier == expected
    assert standard.identifier == expected
    assert enhanced.identifier == expected


def test_indexed_name():
    assert metrics.indexed_name is None
    assert light.indexed_name, 'Kitchin J.'
    assert standard.indexed_name, 'Kitchin J.'
    assert enhanced.indexed_name, 'Kitchin J.'


def test_initials():
    assert metrics.initials is None
    assert light.initials is None
    assert standard.initials, 'J.R.'
    assert enhanced.initials, 'J.R.'


def test_name_variants():
    assert metrics.name_variants is None
    assert light.name_variants is None
    expected = "<class 'pybliometrics.scopus.author_retrieval.Variant'>"
    for a in (standard, enhanced):
        received = a.name_variants
        assert isinstance(received, list)
        assert len(received) > 0
        assert str(type(received[0])) == expected
        assert isinstance(received[0].doc_count, int)


def test_orcid():
    assert metrics.orcid is None
    assert light.orcid, '0000-0003-2625-9232'
    assert standard.orcid, '0000-0003-2625-9232'
    assert enhanced.orcid, '0000-0003-2625-9232'


def test_publication_range():
    assert metrics.publication_range is None
    for a in (standard, enhanced, light):
        assert a.publication_range[0], 1995
        assert a.publication_range[1] >= 2021


def test_scopus_author_link():
    assert metrics.scopus_author_link is None
    expected = 'http://api.elsevier.com/content/author/author_id/7004212771'
    assert light.scopus_author_link == expected
    assert standard.scopus_author_link == expected
    assert enhanced.scopus_author_link == expected


def test_search_link():
    assert metrics.search_link is None
    expected = 'http://api.elsevier.com/content/search/scopus?query='\
               'au-id%287004212771%29'
    assert light.search_link == expected
    assert standard.search_link == expected
    assert enhanced.search_link == expected


def test_self_link():
    assert metrics.self_link is None
    expected = 'https://www.scopus.com/authid/detail.uri?partnerID=HzOxMe3b&'\
               'authorId=7004212771&origin=inward'
    assert light.self_link == expected
    assert standard.self_link == expected
    assert enhanced.self_link == expected


def test_status():
    assert metrics.status is None
    assert light.status is None
    expected = "update"
    assert standard.status == expected
    assert enhanced.status == expected


def test_subject_areas():
    assert metrics.subject_areas is None
    assert light.subject_areas is None
    expected = "<class 'pybliometrics.scopus.author_retrieval.Subjectarea'>"
    for a in (standard, enhanced):
        received = a.subject_areas
        assert isinstance(received, list)
        assert len(received) > 0
        assert str(type(received[0])) == expected


def test_surname():
    assert metrics.surname is None
    assert light.surname is None
    assert standard.surname, 'Kitchin'
    assert enhanced.surname, 'Kitchin'


def test_url():
    expected = 'http://api.elsevier.com/content/author/author_id/7004212771'
    assert metrics.url == expected
    assert light.url == expected
    assert standard.url == expected
    assert enhanced.url == expected



from pybliometrics.scival import init, InstitutionLookupMetrics
from pybliometrics.scival.institution_lookup_metrics import Institution

init()

# Test cases with actual institution IDs from the response examples
single_institution_all = InstitutionLookupMetrics("505023", by_year=False, refresh=30)
multiple_institutions_all = InstitutionLookupMetrics([309054, 309086], by_year=True, refresh=30)
empty_metrics = InstitutionLookupMetrics("0000000")


# Auxiliary function to check if a MetricData namedtuple has all required fields
def has_all_fields(metric_data):
    """Check if the metric data has all required fields."""
    required_fields = ['entity_id', 'entity_name', 'metric',
                       'year', 'value', 'percentage',
                       'threshold']
    return all(hasattr(metric_data, field) for field in required_fields)


def test_academic_corporate_collaboration():
    """Test AcademicCorporateCollaboration property for all test cases."""
    result = single_institution_all.AcademicCorporateCollaboration
    assert has_all_fields(result[0])
    assert result[0].entity_id == 505023
    assert result[0].entity_name == 'Universidad Nacional Autónoma de México'
    assert result[0].metric == 'Academic-corporate collaboration'
    assert result[0].year == 'all'
    assert result[0].value >= 900
    assert result[0].percentage > 2
    assert result[0].threshold is None

    result_multi = multiple_institutions_all.AcademicCorporateCollaboration
    assert has_all_fields(result_multi[0])
    assert result_multi[0].entity_id == 309054
    assert result_multi[0].entity_name == 'Technical University of Munich'
    assert result_multi[0].metric == 'Academic-corporate collaboration'
    assert result_multi[0].year == '2024'
    assert result_multi[0].value >= 1000
    assert result_multi[0].percentage > 9
    assert result_multi[0].threshold is None


def test_academic_corporate_collaboration_impact():
    """Test AcademicCorporateCollaborationImpact property for all test cases."""
    result = single_institution_all.AcademicCorporateCollaborationImpact
    assert has_all_fields(result[0])


def test_institutions():
    """Test the institutions property for all test cases."""

    # Test single institution
    institutions = single_institution_all.institutions
    assert len(institutions) == 1
    expected_institution = Institution(id=505023,
                                        name='Universidad Nacional Autónoma de México',
                                        uri='Institution/505023')
    assert institutions[0] == expected_institution

    # Test multiple institutions
    institutions_multi = multiple_institutions_all.institutions
    assert len(institutions_multi) == 2

    expected_institution_1 = Institution(id=309054,
                                            name='Technical University of Munich',
                                            uri='Institution/309054')
    assert institutions_multi[0] == expected_institution_1

    # Test empty metrics
    assert empty_metrics.institutions is None


def test_citation_count():
    """Test CitationCount property for all test cases."""
    result = single_institution_all.CitationCount
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.CitationCount
    assert has_all_fields(result_multi[0])


def test_citations_per_publication():
    """Test CitationsPerPublication property for all test cases."""
    result = single_institution_all.CitationsPerPublication
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.CitationsPerPublication
    assert has_all_fields(result_multi[0])


def test_cited_publications():
    """Test CitedPublications property for all test cases."""
    result = single_institution_all.CitedPublications
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.CitedPublications
    assert has_all_fields(result_multi[0])


def test_collaboration():
    """Test Collaboration property for all test cases."""
    result = single_institution_all.Collaboration
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.Collaboration
    assert has_all_fields(result_multi[0])


def test_collaboration_impact():
    """Test CollaborationImpact property for all test cases."""
    result = single_institution_all.CollaborationImpact
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.CollaborationImpact
    assert has_all_fields(result_multi[0])


def test_empty_metrics():
    """Test handling of empty metrics."""
    assert empty_metrics.institutions is None 
    assert empty_metrics.CitationCount is None
    assert empty_metrics.CitationsPerPublication is None
    assert empty_metrics.CitedPublications is None
    assert empty_metrics.Collaboration is None
    assert empty_metrics.CollaborationImpact is None


def test_field_weighted_citation_impact():
    """Test FieldWeightedCitationImpact property for all test cases."""
    result = single_institution_all.FieldWeightedCitationImpact
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.FieldWeightedCitationImpact
    assert has_all_fields(result_multi[0])


def test_outputs_in_top_citation_percentiles():
    """Test OutputsInTopCitationPercentiles property for all test cases."""
    result = single_institution_all.OutputsInTopCitationPercentiles
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.OutputsInTopCitationPercentiles
    
    assert has_all_fields(result_multi[0])


def test_publications_in_top_journal_percentiles():
    """Test PublicationsInTopJournalPercentiles property for all test cases."""
    result = single_institution_all.PublicationsInTopJournalPercentiles
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.PublicationsInTopJournalPercentiles
    assert has_all_fields(result_multi[0])


def test_scholarly_output():
    """Test ScholarlyOutput property for all test cases."""
    result = single_institution_all.ScholarlyOutput
    assert has_all_fields(result[0])

    result_multi = multiple_institutions_all.ScholarlyOutput
    assert has_all_fields(result_multi[0])


def test_str_representation():
    """Test the string representation of InstitutionLookupMetrics objects."""
    # Test single institution
    str_single = str(single_institution_all)
    expected_str = "InstitutionLookupMetrics for 1 institution(s):\n- Universidad Nacional Autónoma de México (ID: 505023)"
    assert str_single == expected_str

    # Test multiple institutions
    str_multi = str(multiple_institutions_all)
    expected_str_multi = "InstitutionLookupMetrics for 2 institution(s):\n- Technical University of Munich (ID: 309054)\n- Ludwig Maximilian University of Munich (ID: 309086)"
    assert str_multi == expected_str_multi

    # Test empty metrics
    str_empty = str(empty_metrics)
    assert str_empty == "No institutions found"



from pathlib import Path

# Location of configuration file with legacy support
config_path_options = [
    Path.home() / ".scopus" / "config.ini",
    Path.home() / ".pybliometrics" / "Scopus" / "config.ini",
    Path.home() / ".config" / "pybliometrics.cfg"  # default
]
CONFIG_FILE = next((path for path in config_path_options if path.exists()),
                   config_path_options[-1])

# Location of cache with legacy support
cache_path_options = [
    Path.home() / ".scopus",
    Path.home() / ".pybliometrics",
    Path.home() / ".cache" / "pybliometrics"  # default
]
CACHE_PATH = next((path for path in cache_path_options if path.exists()),
                 cache_path_options[-1])

DEFAULT_PATHS = {
    'AbstractRetrieval': CACHE_PATH / "Scopus" / 'abstract_retrieval',
    'AffiliationRetrieval': CACHE_PATH / "Scopus" / 'affiliation_retrieval',
    'AffiliationSearch': CACHE_PATH / "Scopus" / 'affiliation_search',
    'AuthorRetrieval': CACHE_PATH / "Scopus" / 'author_retrieval',
    'AuthorSearch': CACHE_PATH / "Scopus" / 'author_search',
    'ArticleEntitlement': CACHE_PATH / "ScienceDirect" / 'article_entitlement',
    'ArticleMetadata': CACHE_PATH / "ScienceDirect" / 'article_metadata',
    'ArticleRetrieval': CACHE_PATH / "ScienceDirect" / 'article_retrieval',
    'NonserialTitle': CACHE_PATH / "ScienceDirect" / 'nonserial_title',
    'CitationOverview': CACHE_PATH / "Scopus" / 'citation_overview',
    'ObjectMetadata': CACHE_PATH / "ScienceDirect" / 'object_metadata',
    'ObjectRetrieval': CACHE_PATH / "ScienceDirect" / 'object_retrieval',
    'PlumXMetrics': CACHE_PATH / "Scopus" / 'plumx',
    'PublicationLookup': CACHE_PATH / "Scival" / "publication_lookup",
    'AuthorMetrics': CACHE_PATH / "Scival" / "author_metrics",
    'InstitutionLookupMetrics': CACHE_PATH / "Scival" / "institution_metrics",
    'TopicLookupMetrics': CACHE_PATH / "Scival" / "topic_metrics",
    'ScDirSubjectClassifications': CACHE_PATH / "ScienceDirect" / 'subject_classification',
    'ScienceDirectSearch': CACHE_PATH / "ScienceDirect" / 'science_direct_search',
    'ScopusSearch': CACHE_PATH / "Scopus" / 'scopus_search',
    'SerialTitleSearch': CACHE_PATH / "Scopus" / 'serial_search',
    'SerialTitleISSN': CACHE_PATH / "Scopus" / 'serial_title',
    'SubjectClassifications': CACHE_PATH / "Scopus" / 'subject_classification',
}

# URLs for all classes
RETRIEVAL_BASE = 'https://api.elsevier.com/content/'
SEARCH_BASE = 'https://api.elsevier.com/content/search/'
SCIVAL_BASE = 'https://api.elsevier.com/analytics/scival/'
URLS = {
    'AbstractRetrieval': RETRIEVAL_BASE + 'abstract/',
    'ArticleEntitlement': RETRIEVAL_BASE + 'article/entitlement/',
    'ArticleMetadata': RETRIEVAL_BASE + 'metadata/article/',
    'ArticleRetrieval': RETRIEVAL_BASE + 'article/',
    'AffiliationRetrieval': RETRIEVAL_BASE + 'affiliation/affiliation_id/',
    'AffiliationSearch': SEARCH_BASE + 'affiliation',
    'AuthorRetrieval': RETRIEVAL_BASE + 'author/author_id/',
    'AuthorSearch': SEARCH_BASE + 'author',
    'CitationOverview': RETRIEVAL_BASE + 'abstract/citations/',
    'NonserialTitle': RETRIEVAL_BASE + 'nonserial/title/isbn/',
    'ObjectMetadata': RETRIEVAL_BASE + 'object/',
    'ObjectRetrieval': RETRIEVAL_BASE + 'object/',
    'PublicationLookup': SCIVAL_BASE + 'publication/',
    'AuthorMetrics': SCIVAL_BASE + 'author/metrics/',
    'InstitutionLookupMetrics': SCIVAL_BASE + 'institution/metrics/',
    'TopicLookupMetrics': SCIVAL_BASE + 'topic/metrics/',
    'PlumXMetrics': 'https://api.elsevier.com/analytics/plumx/',
    'ScDirSubjectClassifications': RETRIEVAL_BASE + 'subject/scidir/',
    'ScienceDirectSearch': SEARCH_BASE + 'sciencedirect/',
    'ScopusSearch': SEARCH_BASE + 'scopus',
    'SerialTitleSearch': RETRIEVAL_BASE + 'serial/title',
    'SerialTitleISSN': RETRIEVAL_BASE + 'serial/title/issn/',
    'SubjectClassifications': RETRIEVAL_BASE + 'subject/scopus',
}

# Valid views for all classes
VIEWS = {
    "AbstractRetrieval": ["META", "META_ABS", "FULL", "REF", "ENTITLED"],
    "AffiliationRetrieval": ["LIGHT", "STANDARD", "ENTITLED"],
    "AffiliationSearch": ["STANDARD"],
    "ArticleEntitlement": ["FULL"],
    "ArticleMetadata": ["STANDARD", "COMPLETE"],
    "ArticleRetrieval": ["META", "META_ABS", "META_ABS_REF", "FULL", "ENTITLED"],
    "AuthorRetrieval": ["LIGHT", "STANDARD", "ENHANCED", "METRICS", "ENTITLED"],
    "AuthorSearch": ["STANDARD"],
    "CitationOverview": ["STANDARD"],
    "NonserialTitle": ["STANDARD"],
    "PlumXMetrics": ["ENHANCED"],
    "ScDirSubjectClassifications": [''],
    "ScienceDirectSearch": ["STANDARD"],
    "ScopusSearch": ["STANDARD", "COMPLETE"],
    "SerialTitleSearch": ["STANDARD", "ENHANCED", "CITESCORE"],
    "SerialTitleISSN": ["STANDARD", "ENHANCED", "CITESCORE"],
    "SubjectClassifications": [''],
    "ObjectMetadata": ["META"],
    "ObjectRetrieval": [""]
}

# SciVal Metrics
SCIVAL_METRICS = {
    "AuthorMetrics": {
        "byYear": [
            "AcademicCorporateCollaboration",
            "AcademicCorporateCollaborationImpact",
            "Collaboration",
            "CitationCount",
            "CitationsPerPublication",
            "CollaborationImpact",
            "CitedPublications",
            "FieldWeightedCitationImpact",
            "ScholarlyOutput",
            "PublicationsInTopJournalPercentiles",
            "OutputsInTopCitationPercentiles"
        ],
        "notByYear": [
            "HIndices"
        ]
    },
    "TopicLookupMetrics": {
        "byYear": [
            "AuthorCount",
            "CitationCount",
            "FieldWeightedCitationImpact",
            "InstitutionCount",
            "ScholarlyOutput",
            "TopCitedPublications",
        ],
        "notByYear": [
            "CorePapers",
            "MostRecentlyPublishedPapers",
            "RelatedTopics",
            "TopAuthors",
            "TopInstitutions",
            "TopJournals",
            "TopKeywords"
            ]
    },
    "InstitutionLookupMetrics": {
        "byYear": [
            "AcademicCorporateCollaboration",
            "AcademicCorporateCollaborationImpact",
            "Collaboration",
            "CitationCount",
            "CitationsPerPublication",
            "CollaborationImpact",
            "CitedPublications",
            "FieldWeightedCitationImpact",
            "ScholarlyOutput",
            "PublicationsInTopJournalPercentiles",
            "OutputsInTopCitationPercentiles"
        ],
        "notByYear": []
    }
}

# APIs whose URL needs an id_type
APIS_WITH_ID_TYPE = {"AbstractRetrieval",
                     "PlumXMetrics",
                     "ArticleRetrieval",
                     "ArticleEntitlement",
                     "ObjectMetadata",
                     "ObjectRetrieval"}

# APIs that do not require an ID in the URL
APIS_NO_ID_IN_URL = {"AuthorMetrics", "InstitutionLookupMetrics", "TopicLookupMetrics"}

# Item per page limits for all classes
COUNTS = {
    "AffiliationSearch": {"STANDARD": 200},
    "AuthorSearch": {"STANDARD": 200},
    "ArticleMetadata": {"STANDARD": 200, "COMPLETE": 25},
    "ScDirSubjectClassifications": {"": 200},
    "ScienceDirectSearch": {"STANDARD": 100},
    "ScopusSearch": {"STANDARD": 200, "COMPLETE": 25},
    "SerialTitleSearch": {"STANDARD": 200, "ENHANCED": 200, "CITESCORE": 200},
    "SubjectClassifications": {"": 200}
}

# Throttling limits (in queries per second) // 0 = no limit
RATELIMITS = {
    'AbstractRetrieval': 9,
    'AffiliationRetrieval': 9,
    'AffiliationSearch': 6,
    'ArticleEntitlement': 0,
    'ArticleMetadata': 6,
    'ArticleRetrieval': 10,
    'AuthorMetrics': 6,
    'TopicLookupMetrics': 6,
    'AuthorRetrieval': 3,
    'AuthorSearch': 2,
    'CitationOverview': 4,
    "InstitutionLookupMetrics": 6,
    'NonserialTitle': 6,
    'ObjectMetadata': 0,
    'ObjectRetrieval': 0,
    'PlumXMetrics': 6,
    'PublicationLookup': 6,
    'ScDirSubjectClassifications': 0,
    'ScienceDirectSearch': 2,
    'ScopusSearch': 9,
    'SerialTitleSearch': 6,
    'SerialTitleISSN': 6,
    'SubjectClassifications': 0
}

# Other API restrictions
SEARCH_MAX_ENTRIES = 5_000



"""Utility functions to parse and extract metrics data from JSON responses."""
from typing import NamedTuple

from pybliometrics.utils import make_int_if_possible

class MetricData(NamedTuple):
    """Global namedtuple for all metric data with default values."""
    entity_id: int | None
    entity_name: str | None
    metric: str | None
    year: str | None = "all"
    value: int | float | None = None
    percentage: float | None = None
    threshold: int | None = None


def extract_metric_data(json_data, metric_type: str, by_year: bool, entity_type: str):
    """Helper function to extract metric data for a specific metric type.
    
    Parameters
    ----------
    json_data : dict
        The JSON response from the API
    metric_type : str
        The metric type to extract
    by_year : bool, optional
        Whether the data is broken down by year
    entity_type : str, optional
        The type of entity ("author", "institution", or "topic")
        
    Returns
    -------
    list or None
        List of MetricData namedtuples or None if no data found
    """
    out = []

    # Get results from JSON data
    if isinstance(json_data, dict):
        results = json_data.get('results', [])
    else:
        results = []

    for result in results:
        entity_id, entity_name = extract_entity_info(result, entity_type)
        metric_data = find_metric_data(result, metric_type)

        if not metric_data:
            continue

        # Process metric data using unified approach
        metric_items = process_metric(metric_data, entity_id, entity_name, metric_type, by_year)
        if metric_items:
            out.extend(metric_items)

    return out or None

def extract_metric_lists(json_data, metric_type: str, entity_type: str) -> list:
    out = []

    # Get results from JSON data
    if isinstance(json_data, dict):
        results = json_data.get('results', [])
    else:
        results = []

    for result in results:
        entity_id, entity_name = extract_entity_info(result, entity_type)
        metric_data = find_metric_data(result, metric_type)

        if not metric_data:
            continue

        metric_values = metric_data.get('values', [])
        out.append({
            "entity_id": entity_id,
            "entity_name": entity_name,
            "values": metric_values
        })
    return out


def extract_entity_info(result: dict, entity_type: str) -> tuple:
    """Extract entity ID and name from a result.
    
    Parameters
    ----------
    result : dict
        Entity result from API response
    entity_type : str
        The type of entity ("author", "institution", or "topic")
        
    Returns
    -------
    tuple
        (entity_id, entity_name)
    """
    entity_data = result.get(entity_type, {})
    entity_id = make_int_if_possible(entity_data.get('id'))
    entity_name = entity_data.get('name')
    return entity_id, entity_name


def find_metric_data(result: dict, metric_type: str):
    """Find specific metric data in the metrics list.
    
    Parameters
    ----------
    result : dict
        Author result from API response
    metric_type : str
        The metric type to find
        
    Returns
    -------
    dict or None
        The metric data or None if not found
    """
    metrics = result.get('metrics', [])
    for metric in metrics:
        if metric.get('metricType') == metric_type:
            return metric
    return None


def process_metric(metric_data: dict, entity_id: int, entity_name: str, metric_type: str, by_year: bool = False):
    """Unified function to process all metric types.
    
    Parameters
    ----------
    metric_data : dict
        The metric data from API response
    entity_id : int
        Entity ID (author or institution)
    entity_name : str
        Entity name (author or institution)
    metric_type : str
        The metric type
    by_year : bool, optional
        Whether the data is broken down by year
        
    Returns
    -------
    list or None
        List of MetricData namedtuples or None if no data
    """
    out = []

    # Normalize all metrics to have a 'values' structure
    if 'values' in metric_data:
        # Already has multiple values (collaboration/threshold metrics)
        values_list = metric_data['values']
    else:
        # Simple metric - wrap in a list to make it uniform
        values_list = [metric_data]

    # Process all value items uniformly
    for value_item in values_list:
        # Extract type-specific information - just try to get them, default to None
        collab_type = value_item.get('collabType')
        threshold = value_item.get('threshold')

        # Normalize data structure: convert single values to year-keyed dictionaries
        if by_year:
            value_data = value_item.get('valueByYear', {})
            percentage_data = value_item.get('percentageByYear', {})
        else:
            value_data = {"all": value_item.get('value')}
            percentage_data = {"all": value_item.get('percentage')}

        # Process all years uniformly
        for year in value_data.keys():
            # For nested metrics (like Collaboration), metric is the specific type (collabType)
            # For simple metrics (like CitationCount), metric is the metric_type itself
            metric_name = collab_type or value_item.get('indexType') or value_item.get('impactType') or metric_type

            new = MetricData(
                entity_id=entity_id,
                entity_name=entity_name,
                metric=metric_name,
                year=str(year),
                value=value_data.get(year),
                percentage=percentage_data.get(year),
                threshold=threshold
            )
            out.append(new)

    return out if out else None



"""Test pybliometrics.sciencedirect.SerialTitleISSN()"""

from pybliometrics.sciencedirect import SerialTitleISSN, init
from pybliometrics.scopus.serial_title_issn import Citescoreinfolist, Subjectarea, Yearlydata

init()

gene = SerialTitleISSN('03781119', view='ENHANCED', refresh=30)

def test_module():
    assert gene.__module__  == 'pybliometrics.scopus.serial_title_issn'


def test_deprecated_class():
    from pytest import deprecated_call
    from pybliometrics.sciencedirect import SerialTitle

    with deprecated_call():
        _ = SerialTitle('03781119', refresh=30)


def test_aggregation_type():
    assert gene.aggregation_type == "journal"


def test_citescoreyearinfolist():
    # Test Gene journal
    expected_named_tuple = [Citescoreinfolist(year=2024, citescore=5.1),
                            Citescoreinfolist(year=2025, citescore=4.5)]
    assert gene.citescoreyearinfolist == expected_named_tuple


def test_eissn():
    assert gene.eissn == "1879-0038"


def test_issn():
    assert gene.issn == "0378-1119"


def test_oaallowsauthorpaid():
    assert gene.oaallowsauthorpaid is True


def test_openaccess():
    assert gene.openaccess == 0


def test_openaccessstartdate():
    assert gene.openaccessstartdate is None


def test_openaccesstype():
    assert gene.openaccesstype == 'None'


def test_openaccessarticle():
    assert gene.openaccessarticle is False


def test_openarchivearticle():
    assert gene.openarchivearticle is False


def test_openaccesssponsorname():
    assert gene.openaccesssponsorname is None


def test_openaccessuserlicense():
    assert gene.openaccessuserlicense is None


def test_publisher():
    assert gene.publisher == 'Elsevier B.V.'


def test_scopus_source_link():
    expected1 = 'https://www.scopus.com/source/sourceInfo.url?sourceId=15636'
    assert gene.scopus_source_link == expected1


def test_self_link():
    expected1 = 'https://api.elsevier.com/content/serial/title/issn/03781119'
    assert gene.self_link == expected1


def test_sjrlist():
    assert gene.sjrlist == [(2024, 0.682)]


def test_sniplist():
    assert gene.sniplist == [(2024, 0.716)]


def test_source_id():
    assert gene.source_id == 15636


def test_subject_area():
    expected1 = [
        Subjectarea(area='Genetics', abbreviation='BIOC', code=1311)
    ]
    assert gene.subject_area == expected1


def test_title():
    assert gene.title == 'Gene'


def test_yearly_data():
    assert isinstance(gene.yearly_data, list)
    assert len(gene.yearly_data) == 30

    expected1_2023 = Yearlydata(year=2023, publicationcount=654, revpercent=8.56,
        zerocitessce=64, zerocitespercentsce=9.785932721712538,
        citecountsce=32573)
    assert gene.yearly_data[27] == expected1_2023



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import make_int_if_possible
from pybliometrics.utils.constants import SCIVAL_METRICS
from pybliometrics.utils.parse_metrics import extract_metric_data, MetricData


class Institution(NamedTuple):
    """Named tuple representing an institution."""
    id: int | None
    name: str | None
    uri: str | None


class InstitutionLookupMetrics(Retrieval):
    @property  
    def AcademicCorporateCollaboration(self) -> list[MetricData] | None:
        """Academic corporate collaboration metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'AcademicCorporateCollaboration', self._by_year, "institution")

    @property
    def AcademicCorporateCollaborationImpact(self) -> list[MetricData] | None:
        """Academic corporate collaboration impact metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'AcademicCorporateCollaborationImpact', self._by_year, "institution")

    @property
    def institutions(self) -> list[Institution] | None:
        """A list of namedtuples representing institutions and their basic info
        in the form `(id, name, uri)`.
        """
        out = []

        # Handle both dict and direct access to results
        if isinstance(self._json, dict):
            results = self._json.get('results', [])
        else:
            results = []

        for result in results:
            institution_data = result.get('institution', {})
            new = Institution(
                id=make_int_if_possible(institution_data.get('id')),
                name=institution_data.get('name'),
                uri=institution_data.get('uri')
            )
            out.append(new)
        return out or None

    @property
    def CitationCount(self) -> list[MetricData] | None:
        """Citation count metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitationCount', self._by_year, "institution")

    @property
    def CitationsPerPublication(self) -> list[MetricData] | None:
        """Citations per publication metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitationsPerPublication', self._by_year, "institution")

    @property
    def CitedPublications(self) -> list[MetricData] | None:
        """Cited publications metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitedPublications', self._by_year, "institution")

    @property
    def Collaboration(self) -> list[MetricData] | None:
        """Collaboration metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'Collaboration', self._by_year, "institution")

    @property
    def CollaborationImpact(self) -> list[MetricData] | None:
        """Collaboration impact metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CollaborationImpact', self._by_year, "institution")

    @property
    def FieldWeightedCitationImpact(self) -> list[MetricData] | None:
        """Field weighted citation impact metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'FieldWeightedCitationImpact', self._by_year, "institution")

    @property
    def OutputsInTopCitationPercentiles(self) -> list[MetricData] | None:
        """Outputs in top citation percentiles metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'OutputsInTopCitationPercentiles', self._by_year, "institution")

    @property
    def PublicationsInTopJournalPercentiles(self) -> list[MetricData] | None:
        """Publications in top journal percentiles metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'PublicationsInTopJournalPercentiles', self._by_year, "institution")

    @property
    def ScholarlyOutput(self) -> list[MetricData] | None:
        """Scholarly output metrics for each institution.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'ScholarlyOutput', self._by_year, "institution")

    def __init__(self,
                 institution_ids: str | list,
                 metric_types: str | list | None = None,
                 by_year: bool = False,
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the SciVal's `metrics` endpoint of the `InstitutionLookup API`.

        :param institution_ids: SciVal Institution ID(s). Can be a single ID or comma-separated 
                               string of IDs, or a list of IDs (e.g. `[309054, 309086]`).
        :param metric_types: Metric type(s) to retrieve. Can be a single metric 
                             or comma-separated string, or a list. Available metrics are:
                             AcademicCorporateCollaboration, AcademicCorporateCollaborationImpact,
                             CitationCount, CitedPublications, Collaboration, CollaborationImpact,
                             FieldWeightedCitationImpact, ScholarlyOutput,
                             PublicationsInTopJournalPercentiles, OutputsInTopCitationPercentiles.
                             If not provided, all metrics are retrieved.
        :param by_year: Whether to retrieve metrics broken down by year.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/SciValInstitutionAPI.wadl.
            
        Note:
            All metric properties return lists of MetricData namedtuples with 
            structure: `(entity_id, entity_name, metric, 
            year, value, percentage, threshold)`.

        """
        self._view = ''
        self._refresh = refresh
        self._by_year = by_year

        # Handle institutions parameter
        if isinstance(institution_ids, list):
            institution_ids = ",".join(str(i) for i in institution_ids)

        # Handle metric_types parameter - use all metrics by default
        if metric_types is None:
            metric_types = SCIVAL_METRICS["InstitutionLookupMetrics"]["byYear"]

        if isinstance(metric_types, list):
            metric_types = ",".join(metric_types)

        self._metric_types = metric_types

        # Set up parameters for the API call
        params = {
            'institutionIds': institution_ids,
            'metricTypes': metric_types,
            'byYear': str(by_year).lower(),
            **kwds
        }

        Retrieval.__init__(self, **params, **kwds)

    def __str__(self):
        """Return pretty text version of the institution metrics."""
        institutions = self.institutions or []
        institution_count = len(institutions)

        if institution_count == 0:
            return "No institutions found"
        else:
            s = f"InstitutionLookupMetrics for {institution_count} institution(s):"
            for institution in institutions:
                s += f"\n- {institution.name} (ID: {institution.id})"
            return s



from pybliometrics.utils import *

from pybliometrics.scival.author_metrics import *
from pybliometrics.scival.institution_lookup_metrics import *
from pybliometrics.scival.publication_lookup import *
from pybliometrics.scival.topic_lookup_metrics import *



from typing import NamedTuple

from pybliometrics.superclasses import Search
from pybliometrics.utils import check_field_consistency, chained_get, \
    check_integrity, check_parameter_value, deduplicate, \
    make_search_summary, VIEWS


class Document(NamedTuple):
    """Named tuple representing a document from ScienceDirect Search API."""
    authors: str | None
    first_author: str | None
    doi: str | None
    title: str | None
    link: str | None
    load_date: str | None
    openaccess_status: bool | None
    pii: str | None
    coverDate: str | None
    endingPage: str | None
    publicationName: str | None
    startingPage: str | None
    api_link: str | None
    volume: str | None


class ScienceDirectSearch(Search):
    @property
    def results(self) -> list[Document] | None:
        """A list of namedtuples in the form `(authors first_author doi title link
        load_date openaccess_status pii coverDate endingPage publicationName startingPage
        api_link volume)`.

        Field definitions correspond to the `ScienceDirect Search Views
        <https://dev.elsevier.com/sd_search_views.htmll>`__ and return the
        values as-is, except for `authors` which are joined on `";"`.

        Raises
        ------
        ValueError
            If the elements provided in `integrity_fields` do not match the
            actual field names (listed above).

        Notes
        -----
        The list of authors and the list of affiliations per author are
        deduplicated.
        """
        fields = 'authors first_author doi title link load_date openaccess_status pii '\
            'coverDate endingPage publicationName startingPage api_link volume'
        check_field_consistency(self._integrity, fields)
        # Parse elements one-by-one
        out = []
        for item in self._json:
            # Get authors and create ";" separated string
            authors_list = self._get_authors(item)
            authors_list = deduplicate(authors_list)
            authors = ';'.join(authors_list)
            # Get links
            links_found = item.get('link')
            links = {'api_link': None, 'scidir': None}
            for link in links_found:
                if link.get('@ref') == 'self':
                    links['api_link'] = link.get('@href')
                elif link.get('@ref') == 'scidir':
                    links['scidir'] = link.get('@href')
            # Get doi
            doi = item.get("prism:doi") or item.get("dc:identifier")[4:] if item.get("dc:identifier") else None
            new = Document(
                authors=authors,
                first_author=item.get('dc:creator'),
                doi=doi,
                title=item.get("dc:title"),
                link=links["scidir"],
                load_date=item.get("load-date"),
                openaccess_status=item.get("openaccess"),
                pii=item.get("pii"),
                coverDate=item.get("prism:coverDate"),
                endingPage=item.get("prism:endingPage"),
                publicationName=item.get("prism:publicationName"),
                startingPage=item.get("prism:startingPage"),
                api_link=links["api_link"] or item.get("prism:url"),
                volume=item.get("prism:volume")
            )
            out.append(new)
        check_integrity(out, self._integrity, self._action)
        return out or None

    def __init__(self,
                 query: str,
                 refresh: bool | int = False,
                 view: str | None = None,
                 verbose: bool = False,
                 download: bool = True,
                 integrity_fields: list[str] | tuple[str, ...] | None = None,
                 integrity_action: str = "raise",
                 subscriber: bool = True,
                 **kwds: str
                 ) -> None:
        """Interaction with the ScienceDirect Search API. This represents a search against the
        ScienceDirect cluster, which contains serial/nonserial full-text articles. Note that this API
        replicates the search experience on `ScienceDirect <www.sciencedirect.com>`__.

        :param query: A string of the query as used in the `ScienceDirect Search <https://dev.elsevier.com/tecdoc_sdsearch_migration.html>`__.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: Which view to use for the query, see `the documentation <https://dev.elsevier.com/sd_search_views.html>`__.
                     Allowed values: `STANDARD`.
        :param verbose: Whether to print a download progress bar.
        :param download: Whether to download results (if they have not been
                         cached).
        :param integrity_fields: A list or tuple with the names of fields whose completeness should
                                 be checked.  `ArticleMetadata` will perform the
                                 action specified in `integrity_action` if
                                 elements in these fields are missing.  This
                                 helps to avoid idiosynchratically missing
                                 elements that should always be present
                                 (e.g., doi or authors).
        :param integrity_action: What to do in case integrity of provided fields
                                 cannot be verified.  Possible actions:
                                 - `"raise"`: Raise an `AttributeError`
                                 - `"warn"`: Raise a `UserWarning`
        :param subscriber: Whether you access ScienceDirect with a subscription or not.
                           For subscribers, ScienceDirect's cursor navigation will be
                           used.  Sets the number of entries in each query
                           iteration to the maximum number allowed by the
                           corresponding view.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the `API specification <https://dev.elsevier.com/documentation/ArticleMetadataAPI.wadl>`__.

        Raises
        ------
        ScopusQueryError
            For non-subscribers, if the number of search results exceeds 5000.

        ValueError
            If any of the parameters `integrity_action`, `refresh` or `view`
            is not one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{fname}`,
        where `path` is specified in your configuration file and `fname` is
        the md5-hashed version of `query`.

        The ScienceDirect Search API V2 has two available interfaces: `PUT` and `GET`. This library uses the
        `GET` interface.
        """
        # Check view or set to default
        if view:
            check_parameter_value(view, VIEWS['ScienceDirectSearch'], "view")
        else:
            view = "STANDARD"

        allowed = ("warn", "raise")
        check_parameter_value(integrity_action, allowed, "integrity_action")

        # Query
        self._action = integrity_action
        self._integrity = integrity_fields or []
        self._refresh = refresh
        self._query = query
        self._view = view
        Search.__init__(self, query=query, download=download, verbose=verbose, **kwds)

    def __str__(self):
        """Print a summary string."""
        return make_search_summary(self, "document", self.get_dois())

    def get_dois(self):
        """DOIs of retrieved documents."""
        return [d.get("prism:doi") or d.get("dc:identifier")[4:] if d.get("dc:identifier") else None for d in self._json]

    def _get_authors(self, item: dict) -> list:
        """Auxiliary function to get the authors."""
        authors_data = chained_get(item, ['authors', 'author'], [])
        if isinstance(authors_data, list):
            authors_list = [a.get('$') for a in authors_data]
        elif isinstance(authors_data, str):
            authors_list = [authors_data]
        else:
            authors_list = []
        return authors_list



from pybliometrics.scival.publication_lookup import PublicationLookup
from pybliometrics.utils.startup import init

init()

pub1 = PublicationLookup('85036568406')


def test_publication_authors_count():
    assert len(pub1.authors) >= 7


def test_publication_citation_count():
    assert pub1.citation_count > 0


def test_publication_doi():
    assert pub1.doi == "10.1002/anie.201709271"


def test_publication_first_author():
    assert pub1.authors[0].id == 7404861905
    assert pub1.authors[0].name == "Lin, T.-E."
    assert pub1.authors[0].uri == "Author/7404861905"


def test_publication_first_institution():
    assert pub1.institutions[0].id == 217002
    assert pub1.institutions[0].name == "Chang Gung University"
    assert pub1.institutions[0].country == "Taiwan"
    assert pub1.institutions[0].country_code == "TWN"


def test_publication_id():
    assert pub1.id == 85036568406


def test_publication_institutions_count():
    assert len(pub1.institutions) >= 3


def test_publication_sdgs():
    assert len(pub1.sdgs) >= 1
    assert pub1.sdgs[0] == 'SDG 3: Good Health and Well-being'


def test_publication_source_title():
    assert pub1.source_title == 'Angewandte Chemie - International Edition'


def test_publication_type():
    assert pub1.type == "Article"


def test_publication_year():
    assert pub1.publication_year == 2017



from pybliometrics.utils import *

from pybliometrics.scopus.abstract_citation import *
from pybliometrics.scopus.abstract_retrieval import *
from pybliometrics.scopus.affiliation_retrieval import *
from pybliometrics.scopus.affiliation_search import *
from pybliometrics.scopus.author_retrieval import *
from pybliometrics.scopus.author_search import *
from pybliometrics.scopus.plumx_metrics import *
from pybliometrics.scopus.scopus_search import *
from pybliometrics.scopus.serial_title_search import *
from pybliometrics.scopus.serial_title_issn import *
from pybliometrics.scopus.subject_classifications import *



def check_parameter_value(parameter, allowed, name):
    """Raise a ValueError if a parameter value is not in the set of
    allowed values.
    """
    if parameter not in allowed:
        raise ValueError(f"Parameter '{name}' must be one of {', '.join(allowed)}.")



from pybliometrics.utils.checks import *
from pybliometrics.utils.constants import *
from pybliometrics.utils.create_config import *
from pybliometrics.utils.get_content import *
from pybliometrics.utils.parse_content import *
from pybliometrics.utils.parse_metrics import *
from pybliometrics.utils.startup import *



from typing import NamedTuple

from pybliometrics.superclasses import Search
from pybliometrics.utils import check_integrity, check_parameter_value, \
    check_field_consistency, deduplicate, get_freetoread, html_unescape, \
    listify, make_search_summary, VIEWS


class Document(NamedTuple):
    eid: str | None
    doi: str | None
    pii: str | None
    pubmed_id: str | None
    title: str | None
    subtype: str | None
    subtypeDescription: str | None
    creator: str | None
    afid: str | None
    affilname: str | None
    affiliation_city: str | None
    affiliation_country: str | None
    author_count: str | None
    author_names: str | None
    author_ids: str | None
    author_afids: str | None
    coverDate: str | None
    coverDisplayDate: str | None
    publicationName: str | None
    issn: str | None
    source_id: str | None
    eIssn: str | None
    aggregationType: str | None
    volume: str | None
    issueIdentifier: str | None
    article_number: str | None
    pageRange: str | None
    description: str | None
    authkeywords: str | None
    citedby_count: int
    openaccess: int
    freetoread: str | None
    freetoreadLabel: str | None
    fund_acr: str | None
    fund_no: str | None
    fund_sponsor: str | None


class ScopusSearch(Search):
    @property
    def results(self) -> list[Document] | None:
        """A list of namedtuples in the form `(eid doi pii pubmed_id title
        subtype subtypeDescription creator afid affilname affiliation_city
        affiliation_country author_count author_names author_ids author_afids
        coverDate coverDisplayDate publicationName issn source_id eIssn
        aggregationType volume issueIdentifier article_number pageRange
        description authkeywords citedby_count openaccess freetoread
        freetoreadLabel fund_acr fund_no fund_sponsor)`.
        Field definitions correspond to
        https://dev.elsevier.com/guides/ScopusSearchViews.htm and return the
        values as-is, except for `afid`, `affilname`, `affiliation_city`,
        `affiliation_country`, `author_names`, `author_ids` and `author_afids`:
        This information is joined on `";"`.  In case an author has multiple
        affiliations, they are joined on `"-"`
        (e.g. `Author1Aff;Author2Aff1-Author2Aff2`).

        Raises
        ------
        ValueError
            If the elements provided in `integrity_fields` do not match the
            actual field names (listed above).

        Notes
        -----
        The list of authors and the list of affiliations per author are
        deduplicated.

        The Scopus API returns only the first funding information.
        """
        # Initiate namedtuple with ordered list of fields
        fields = 'eid doi pii pubmed_id title subtype subtypeDescription ' \
                 'creator afid affilname affiliation_city ' \
                 'affiliation_country author_count author_names author_ids '\
                 'author_afids coverDate coverDisplayDate publicationName '\
                 'issn source_id eIssn aggregationType volume '\
                 'issueIdentifier article_number pageRange description '\
                 'authkeywords citedby_count openaccess freetoread '\
                 'freetoreadLabel fund_acr fund_no fund_sponsor'
        check_field_consistency(self._integrity, fields)
        # Parse elements one-by-one
        out = []
        for item in self._json:
            info = {}
            # Parse affiliations
            for field, key in [('affilname', 'affilname'),
                               ('afid', 'afid'),
                               ('aff_city', 'affiliation-city'),
                               ('aff_country', 'affiliation-country')]:
                info[field] = _join(item, key, unescape=self.unescape)
            # Parse authors
            try:
                # Deduplicate list of authors
                authors = deduplicate(item['author'])
                # Extract information
                surnames = _replace_none([d['surname'] for d in authors])
                firstnames = _replace_none([d['given-name'] for d in authors])
                info["auth_names"] = ";".join([", ".join([t[0], t[1]]) for t in
                                               zip(surnames, firstnames)])
                info["auth_ids"] = ";".join([d['authid'] for d in authors])
                affs = []
                for auth in authors:
                    aff = listify(deduplicate(auth.get('afid', [])))
                    affs.append('-'.join([d['$'] for d in aff]))
                if [a for a in affs if a]:
                    info["auth_afid"] = ';'.join(affs)
                else:
                    info["auth_afid"] = None
            except KeyError:
                pass
            date = item.get('prism:coverDate')
            if isinstance(date, list):
                date = date[0].get('$')
            freetoread = get_freetoread(item, ["freetoread", "value"])
            freetoreadLabel = get_freetoread(item, ["freetoreadLabel", "value"])
            # Get text fields and unescape
            for key in ['dc:title', 'dc:description', 'authkeywords']:
                value = item.get(key)
                info[key] = html_unescape(str(value)) if (self.unescape and value) else value
            fund_no = item.get('fund-no', '').replace("undefined", "") or None
            new = Document(article_number=item.get('article-number'),
                      title=info.get('dc:title'),
                      fund_no=fund_no,
                      fund_sponsor=item.get('fund-sponsor'),
                      subtype=item.get('subtype'), doi=item.get('prism:doi'),
                      subtypeDescription=item.get('subtypeDescription'),
                      issn=item.get('prism:issn'), creator=item.get('dc:creator'),
                      affilname=info.get("affilname"),
                      author_names=info.get("auth_names"),
                      coverDate=date, volume=item.get('prism:volume'),
                      coverDisplayDate=item.get('prism:coverDisplayDate'),
                      publicationName=item.get('prism:publicationName'),
                      source_id=item.get('source-id'), author_ids=info.get("auth_ids"),
                      aggregationType=item.get('prism:aggregationType'),
                      issueIdentifier=item.get('prism:issueIdentifier'),
                      pageRange=item.get('prism:pageRange'),
                      author_afids=info.get("auth_afid"),
                      affiliation_country=info.get("aff_country"),
                      citedby_count=int(item['citedby-count']),
                      openaccess=int(item['openaccess']),
                      freetoread=freetoread, freetoreadLabel=freetoreadLabel,
                      eIssn=item.get('prism:eIssn'),
                      author_count=item.get('author-count', {}).get('$'),
                      affiliation_city=info.get("aff_city"), afid=info.get("afid"),
                      description=info.get('dc:description'),
                      pii=item.get('pii'),
                      authkeywords=info.get('authkeywords'),
                      eid=item.get('eid'),
                      fund_acr=item.get('fund-acr'), pubmed_id=item.get('pubmed-id'))
            out.append(new)
        # Finalize
        check_integrity(out, self._integrity, self._action)
        return out or None

    def __init__(self,
                 query: str,
                 refresh: bool | int = False,
                 view: str = None,
                 verbose: bool = False,
                 download: bool = True,
                 integrity_fields: list[str] | tuple[str, ...] | None = None,
                 integrity_action: str = "raise",
                 subscriber: bool = True,
                 unescape: bool = True,
                 **kwds: str
                 ) -> None:
        """Interaction with the Scopus Search API.

        :param query: A string of the query as used in the Advanced Search
                     on scopus.com.  All fields except "INDEXTERMS()" and
                     "LIMIT-TO()" work.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: Which view to use for the query, see
                     https://dev.elsevier.com/sc_search_views.html.
                     Allowed values: `STANDARD`, `COMPLETE`.  If `None`, defaults to
                     `COMPLETE` if `subscriber=True` and to `STANDARD` if
                     `subscriber=False`.
        :param verbose: Whether to print a download progress bar.
        :param download: Whether to download results (if they have not been
                         cached).
        :param integrity_fields: Names of fields whose completeness should
                                 be checked.  `ScopusSearch` will perform the
                                 action specified in `integrity_action` if
                                 elements in these fields are missing.  This
                                 helps to avoid idiosynchratically missing
                                 elements that should always be present
                                 (e.g., EID or source ID).
        :param integrity_action: What to do in case integrity of provided fields
                                 cannot be verified.  Possible actions:
                                 - `"raise"`: Raise an `AttributeError`
                                 - `"warn"`: Raise a `UserWarning`
        :param subscriber: Whether you access Scopus with a subscription or not.
                           For subscribers, Scopus's cursor navigation will be
                           used.  Sets the number of entries in each query
                           iteration to the maximum number allowed by the
                           corresponding view.
        :param unescape: Convert named and numeric characters in the `results` to
                         their corresponding Unicode characters.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/ScopusSearchAPI.wadl.

        Raises
        ------
        ScopusQueryError
            For non-subscribers, if the number of search results exceeds 5000.

        ValueError
            If any of the parameters `integrity_action`, `refresh` or `view`
            is not one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{fname}`,
        where `path` is specified in your configuration file and `fname` is
        the md5-hashed version of `query`.
        """
        # Checks
        if view:
            check_parameter_value(view, VIEWS['ScopusSearch'], "view")
        allowed = ("warn", "raise")
        check_parameter_value(integrity_action, allowed, "integrity_action")

        # Parameters
        if not view:
            if subscriber:
                view = "COMPLETE"
            else:
                view = "STANDARD"
        if "cursor" in kwds:
            subscriber = kwds["cursor"]
            kwds.pop("cursor")

        # Query
        self._action = integrity_action
        self._integrity = integrity_fields or []
        self._refresh = refresh
        self._query = query
        self._view = view
        Search.__init__(self, query=query,
                        cursor=subscriber, download=download,
                        verbose=verbose, **kwds)
        self.unescape = unescape

    def __str__(self):
        """Print a summary string."""
        return make_search_summary(self, "document", self.get_eids())

    def get_eids(self):
        """EIDs of retrieved documents."""
        return [d['eid'] for d in self._json]


def _join(item, key, sep=";", unescape=False):
    """Auxiliary function to join same elements of a list of dictionaries if
    the elements are not None.
    """
    try:
        string = sep.join([d[key] or "" for d in item["affiliation"]])
        return html_unescape(string) if unescape else string
    except (KeyError, TypeError):
        return None


def _replace_none(lst, repl=""):
    """Auxiliary function to replace None's with another value."""
    return [repl if v is None else v for v in lst]



from typing import NamedTuple
from functools import reduce
from html import unescape
from warnings import warn


class Affiliation(NamedTuple):
    """Named tuple representing affiliation information."""
    id: int | None
    parent: int | None
    type: str | None
    relationship: str | None
    afdispname: str | None
    preferred_name: str | None
    parent_preferred_name: str | None
    country_code: str | None
    country: str | None
    address_part: str | None
    city: str | None
    state: str | None
    postal_code: str | None
    org_domain: str | None
    org_URL: str | None


def filter_digits(s):
    """Helper function to remove non-digits characters from a string."""
    return "".join(filter(str.isdigit, s))


def chained_get(container, path, default=None):
    """Helper function to perform a series of .get() methods on a dictionary
    or return the `default`.

    Parameters
    ----------
    container : dict
        The dictionary on which the .get() methods should be performed.

    path : list or tuple
        The list of keys that should be searched for.

    default : any (optional, default=None)
        The object type that should be returned if the search yields
        no result.
    """
    # Obtain value via reduce
    try:
        return reduce(lambda c, k: c.get(k, default), path, container)
    except (AttributeError, TypeError):
        return default


def check_integrity(tuples, fields, action):
    """Check integrity of specific fields in a list of tuples and perfom
    provided action.
    """
    for field in fields:
        elements = [getattr(e, field) for e in tuples]
        if None not in elements:
            continue
        msg = "Parsed information doesn't pass integrity check because of "\
              f"incomplete information in field '{field}'"
        if action == "raise":
            raise AttributeError(msg)
        elif action == "warn":
            warn(msg)


def check_field_consistency(needles, haystack):
    """Raise ValueError if elements of a list are not present in a string."""
    wrong = set(needles) - set(haystack.split())
    if wrong:
        msg = f"Element(s) '{', '.join(sorted(wrong))}' not allowed in "\
              "parameter integrity_fields"
        raise ValueError(msg)


def deduplicate(lst):
    """Auxiliary function to deduplicate a list while preserving its order."""
    new = reduce(lambda x, y: x + y if y[0] not in x else x,
                 map(lambda x: [x], lst),
                 [])
    return new

def get_and_aggregate_subjects(fields):
    """Get and aggregate subject areas from Scopus AuthorSearch."""
    frequencies = {}
    for field_d in fields:
        abbrev = field_d.get('@abbrev', '')
        freq_str = field_d.get('@frequency', '')
        frequency = int(freq_str) if freq_str.isdigit() else 0
        if abbrev in frequencies:
            frequencies[abbrev] += frequency
        else:
            frequencies[abbrev] = frequency
    return frequencies


def get_id(s, integer=True):
    """Helper function to return the Scopus ID at a fixed position."""
    path = ['coredata', 'dc:identifier']
    try:
        return int(chained_get(s, path, "").split(':')[-1])
    except ValueError:
        return None


def get_freetoread(item, path):
    """Helper function to return freetoread information from search results."""
    text = chained_get(item, path)
    try:
        text = " ".join([x['$'] for x in text])
    except TypeError:
        pass
    return text


def get_link(dct, idx, path=['coredata', 'link']):
    """Helper function to return the link at position `idx` from coredata."""
    links = chained_get(dct, path, [{}])
    try:
        return links[idx].get('@href')
    except IndexError:
        return None


def html_unescape(s: str | None) -> str | None:
    """Convert s to Unicode characters if possible."""
    return unescape(s) if s else None


def listify(element):
    """Helper function to turn an element into a list if it isn't a list yet.
    """
    if isinstance(element, list):
        return element
    else:
        return [element]


def list_authors(lst):
    """Format a list of authors (Surname, Firstname and Firstname Surname)."""
    authors = ', '.join([' '.join([a.given_name, a.surname]) for a in lst[0:-1]])
    authors += ' and ' + ' '.join([lst[-1].given_name, lst[-1].surname])
    return authors


def make_float_if_possible(val):
    """Attempt a conversion to float type."""
    try:
        return float(val)
    except TypeError:
        return val


def make_int_if_possible(val):
    """Attempt a conversion to int type."""
    try:
        return int(val)
    except TypeError:
        return val


def make_bool_if_possible(val):
    """Attempt a conversion to bool type."""
    if isinstance(val, str):
        return val.lower() == 'true'
    elif isinstance(val, int):
        return bool(val)
    else:
        return val
    

def make_search_summary(self, keyword, results, joiner="\n    "):
    """Create string for str dunder of search classes."""
    date = self.get_cache_file_mdate().split()[0]
    if self._n != 1:
        appendix = "s"
        verb = "have"
    else:
        appendix = ""
        verb = "has"
    s = f"Search '{self._query}' yielded {self._n:,} "\
        f"{keyword}{appendix} as of {date}"
    if results:
        s += ":" + joiner + joiner.join(results)
    elif self._n:
        s += f", which {verb} not been downloaded"
    return s


def parse_affiliation(affs, view) -> list[Affiliation] | None:
    """Helper function to parse list of affiliation-related information."""
    out = []

    if view in ('STANDARD', 'ENHANCED'):
        for item in listify(affs):
            if not item:
                continue
            doc = item.get('ip-doc', {}) or {}
            address = doc.get('address', {}) or {}
            try:
                parent = int(item['@parent'])
            except KeyError:
                parent = None
            new = Affiliation(id=int(item['@affiliation-id']), parent=parent,
                      type=doc.get('@type'), relationship=doc.get('@relationship'),
                      afdispname=doc.get('@afdispname'),
                      preferred_name=doc.get('preferred-name', {}).get('$'),
                      parent_preferred_name=doc.get('parent-preferred-name', {}).get('$'),
                      country_code=address.get('@country'), country=address.get('country'),
                      address_part=address.get("address-part"), city=address.get('city'),
                      postal_code=address.get('postal-code'), state=address.get('state'),
                      org_domain=doc.get('org-domain'), org_URL=doc.get('org-URL'))
            if any(val for val in new):
                out.append(new)
    elif view == 'LIGHT':
        new = Affiliation(id=None, parent=None, type=None, relationship=None,
                  afdispname=None, preferred_name=affs.get('affiliation-name'),
                  parent_preferred_name=None, country_code=None, country=affs.get('affiliation-country'),
                  address_part=None, city=affs.get('affiliation-city'), state=None,
                  postal_code=None, org_domain=None, org_URL=None)
        if any(val for val in new):
            out.append(new)

    return out or None


def parse_date_created(dct):
    """Helper function to parse date-created from profile."""
    date = dct['date-created']
    if date:
        return int(date['@year']), int(date['@month']), int(date['@day'])
    else:
        return None, None, None


def parse_pages(self, unicode=False):
    """Auxiliary function to parse and format page range of a document."""
    if self.pageRange:
        pages = f'pp. {self.pageRange}'
    elif self.startingPage:
        pages = f'pp. {self.startingPage}-{self.endingPage}'
    else:
        pages = '(no pages found)'
    if unicode:
        pages = f'{pages}'
    return pages



"""Module with the ObjectMetadata class."""

from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value, detect_id_type, make_int_if_possible, VIEWS


class Metadata(NamedTuple):
    """Named tuple representing metadata of a document object."""
    eid: str
    filename: str | None
    height: int | None
    mimetype: str | None
    ref: str | None
    size: int | None
    type: str | None
    url: str | None
    width: int | None


class ObjectMetadata(Retrieval):
    """Class to retrieve a the metadata of all objects of a document."""
    @property
    def results(self) -> list[Metadata]:
        """Metadata of the objects in a document. List of namedtuples in the form `eid`, `filename`,
        `height`, `mimetype`, `ref`, `size`, `type`, `url` and `width`.
        """
        refs = chained_get(self._json, ['attachment-metadata-response', 'attachment'], [])
        out = []
        for ref in refs:
            out.append(
                Metadata(
                    eid=ref["eid"],
                    filename=ref.get("filename"),
                    height=make_int_if_possible(ref.get("height")),
                    mimetype=ref.get("mimetype"),
                    ref=ref.get("ref"),
                    size=make_int_if_possible(ref.get("size")),
                    type=ref.get("type"),
                    url=ref.get("prism:url"),
                    width=make_int_if_possible(ref.get("width")),
                )
            )
        return out

    def __init__(self,
                 identifier: int | str,
                 view: str = 'META',
                 id_type: str | None = None,
                 refresh: bool | int = False,
                 **kwds: str
                 ):
        """Class to retrieve the metadata of all objects of a document.

        :param identifier: The indentifier of an article.
        :param view: The view of the object. Allowed value: `META`.
        :param id_type: The type of identifier supplied. Allowed values: `doi`,
                        `pii`, `scopus_id`, `pubmed_id`, `eid`.
        :param refresh: Whether to refresh the cached file if it exists. Default: `False`.
        """
        self.identifier = str(identifier)
        check_parameter_value(view, VIEWS['ObjectMetadata'], "view")

        self.id_type = id_type
        if self.id_type is None:
            self.id_type = detect_id_type(self.identifier)
        else:
            allowed_id_types = ('doi', 'pii', 'scopus_id', 'pubmed_id', 'eid')
            check_parameter_value(id_type, allowed_id_types, "id_type")

        self._view = view
        self._refresh = refresh

        super().__init__(self.identifier, self.id_type, **kwds)

    def __str__(self):
        return (f'Document with {self.id_type} {self.identifier} contains '
                f'{len(self.results)} objects.')



"""Nonserial title class."""

from pybliometrics.superclasses import Retrieval

from pybliometrics.utils import chained_get, check_parameter_value, VIEWS



class NonserialTitle(Retrieval):
    @property
    def aggregation_type(self) -> str:
        """The type of the source."""
        return self._entry['prism:aggregationType']

    @property
    def authors(self) -> str | None:
        """The authors of the book."""
        return self._entry.get('authors')


    @property
    def description(self) -> str | None:
        """The description of the book."""
        return self._entry.get('description')

    @property
    def edition(self) -> str | None:
        """The edition of the book."""
        return self._entry.get('prism:edition')

    @property
    def editors(self) -> str | None:
        """The editors of the book."""
        return self._entry.get('editors')

    @property
    def isbn(self) -> str:
        """The ISBN of the book."""
        return self._entry['prism:isbn']

    @property
    def link_coverimage(self) -> str:
        """The link to the cover image of the book."""
        links = self._entry.get('link')
        return [link['@href'] for link in links if link['@rel'] == 'coverimage'][0]

    @property
    def link_homepage(self) -> str:
        """The link to the homepage of the book."""
        links = self._entry.get('link')
        return [link['@href'] for link in links if link['@rel'] == 'homepage'][0]

    @property
    def link_search(self) -> str:
        """The link to search for the book."""
        links = self._entry.get('link')
        return [link['@href'] for link in links if link['@rel'] == 'search'][0]

    @property
    def publisher_id(self) -> str:
        """The publisher id of the book."""
        return chained_get(self._entry, ['dc:publisher', '@id'])

    @property
    def publisher_name(self) -> str:
        """The publisher of the book."""
        return chained_get(self._entry, ['dc:publisher', '$'])

    @property
    def self_link(self) -> str:
        """URL to the source's API page."""
        return self._entry['prism:url']

    @property
    def title(self) -> str:
        """The title of the book."""
        return self._entry['dc:title']

    def __init__(self,
                 isbn: int | str,
                 view: str = "STANDARD",
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the ScienceDirect Nonserial Title API.

        :param isbn: The ISBN of the book.
        :param view: The view of the file that should be downloaded. Allowed value: "STANDARD".
                     For details see `the documentation <https://dev.elsevier.com/sd_nonserial_title_views.html>`_.
                     Note that although the "BASIC" view is documented, the API does not support it.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the number of days since
                        last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters. Must contain fields and values
                     mentioned in the `API specification <https://dev.elsevier.com/documentation/NonSerialTitleAPI.wadl>`_.

        :raises ValueError: If any of the parameters `refresh` or `view` is not one of the allowed values.

        .. note::
           The directory for cached results is ``{path}/{view}/{source_id}``,
           where `path` is specified in your configuration file.
        """
        # Checks
        check_parameter_value(view, VIEWS['NonserialTitle'], "view")

        self._view = view
        self._refresh = refresh
        self._id = str(isbn)

        # Load json
        Retrieval.__init__(self, identifier=self._id, **kwds)

        # Parse json
        self._json = self._json['nonserial-metadata-response']
        self._entry = self._json['entry'][0]

    def __str__(self) -> str:
        """Return a citation string for the Nonserial Title."""
        if self.authors:
            authors = f"{self.authors}."
        elif self.editors:
            authors = f"{self.editors} (ed.)."
        else:
            authors = ""
        edition = f", {self.edition}." if self.edition else "."
        publisher = f" {self.publisher_name}." if self.publisher_name else "."
        title = f" {self.title}"
        isbn = self.isbn

        return f"{authors}{title}{edition}{publisher} ISBN: {isbn}"



from importlib.metadata import version

__version__ = version("pybliometrics")

__citation__ = 'Rose, Michael E. and John R. Kitchin: "pybliometrics: '\
    'Scriptable bibliometrics using a Python interface to Scopus", SoftwareX '\
    '10 (2019) 100263.'

import pybliometrics.scopus
import pybliometrics.sciencedirect
from pybliometrics.utils.startup import init



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import make_int_if_possible
from pybliometrics.utils.constants import SCIVAL_METRICS
from pybliometrics.utils.parse_metrics import extract_metric_data, MetricData


class Author(NamedTuple):
    """Named tuple representing an author."""
    id: int | None
    name: str | None
    uri: str | None


class AuthorMetrics(Retrieval):
    @property  
    def AcademicCorporateCollaboration(self) -> list[MetricData] | None:
        """Academic corporate collaboration metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'AcademicCorporateCollaboration', self._by_year, "author")

    @property
    def AcademicCorporateCollaborationImpact(self) -> list[MetricData] | None:
        """Academic corporate collaboration impact metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'AcademicCorporateCollaborationImpact', self._by_year, "author")

    @property
    def authors(self) -> list[MetricData] | None:
        """A list of namedtuples representing authors and their basic info
        in the form `(id, name, uri)`.
        """
        out = []

        # Handle both dict and direct access to results
        if isinstance(self._json, dict):
            results = self._json.get('results', [])
        else:
            results = []

        for result in results:
            author_data = result.get('author', {})
            new = Author(
                id=make_int_if_possible(author_data.get('id')),
                name=author_data.get('name'),
                uri=author_data.get('uri')
            )
            out.append(new)
        return out or None

    @property
    def CitationCount(self) -> list[MetricData] | None:
        """Citation count metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitationCount', self._by_year, "author")

    @property
    def CitationsPerPublication(self) -> list[MetricData] | None:
        """Citations per publication metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitationsPerPublication', self._by_year, "author")

    @property
    def CitedPublications(self) -> list[MetricData] | None:
        """Cited publications metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CitedPublications', self._by_year, "author")

    @property
    def Collaboration(self) -> list[MetricData] | None:
        """Collaboration metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'Collaboration', self._by_year, "author")

    @property
    def CollaborationImpact(self) -> list[MetricData] | None:
        """Collaboration impact metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'CollaborationImpact', self._by_year, "author")

    @property
    def FieldWeightedCitationImpact(self) -> list[MetricData] | None:
        """Field weighted citation impact metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'FieldWeightedCitationImpact', self._by_year, "author")

    @property
    def HIndices(self) -> list[MetricData] | None:
        """H-indices metrics for each author (only available when by_year=False).
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'HIndices', self._by_year, "author")

    @property
    def OutputsInTopCitationPercentiles(self) -> list[MetricData] | None:
        """Outputs in top citation percentiles metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'OutputsInTopCitationPercentiles', self._by_year, "author")

    @property
    def PublicationsInTopJournalPercentiles(self) -> list[MetricData] | None:
        """Publications in top journal percentiles metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'PublicationsInTopJournalPercentiles', self._by_year, "author")

    @property
    def ScholarlyOutput(self) -> list[MetricData] | None:
        """Scholarly output metrics for each author.
        Returns list of MetricData namedtuples with structure:
        (entity_id, entity_name, metric, year, value, percentage, threshold).
        """
        return extract_metric_data(self._json, 'ScholarlyOutput', self._by_year, "author")

    def __init__(self,
                 author_ids: str | list,
                 metric_types: str | list | None = None,
                 by_year: bool = False,
                 refresh: bool | int = False,
                 **kwds: str
                 ) -> None:
        """Interaction with the SciVal Author Metrics API.

        :param author_ids: Scopus Author ID(s). Can be a single ID or comma-separated 
                           string of IDs, or a list of IDs (e.g. `[55586732900, 57215631099]`).
        :param metric_types: Metric type(s) to retrieve. Can be a single metric 
                             or comma-separated string, or a list. Available metrics are:
                             AcademicCorporateCollaboration, AcademicCorporateCollaborationImpact,
                             CitationCount, CitedPublications, Collaboration, CollaborationImpact,
                             FieldWeightedCitationImpact, ScholarlyOutput,
                             PublicationsInTopJournalPercentiles, OutputsInTopCitationPercentiles,
                             HIndices.
                             If not provided, all metrics are retrieved.
        :param by_year: Whether to retrieve metrics broken down by year.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/SciValAuthorAPI.wadl.
            
        Note:
            All metric properties return lists of MetricData namedtuples with 
            structure: `(entity_id, entity_name, metric, 
            year, value, percentage, threshold)` which enable concatenation
            of results from different metrics.
        """
        self._view = ''
        self._refresh = refresh
        self._by_year = by_year

        # Handle authors parameter
        if isinstance(author_ids, list):
            author_ids = ",".join(str(a) for a in author_ids)

        # Handle metric_types parameter - use all metrics by default
        if metric_types is None:
            if not by_year:
                metric_types = SCIVAL_METRICS["AuthorMetrics"]["byYear"] + SCIVAL_METRICS["AuthorMetrics"]["notByYear"]
            if by_year:
                metric_types = SCIVAL_METRICS["AuthorMetrics"]["byYear"]

        if isinstance(metric_types, list):
            metric_types = ",".join(metric_types)

        self._metric_types = metric_types

        # Set up parameters for the API call
        params = {
            'authors': author_ids,
            'metricTypes': metric_types,
            'byYear': str(by_year).lower(),
            **kwds
        }

        Retrieval.__init__(self, **params)

    def __str__(self):
        """Return pretty text version of the author metrics."""
        authors = self.authors or []
        author_count = len(authors)

        if author_count == 0:
            return "No authors found"
        else:
            s = f"AuthorMetrics for {author_count} author(s):"
            for author in authors:
                s += f"\n- {author.name} (ID: {author.id})"
            return s



"""Tests for sciencedirect.ArticleRetrieval"""

from pybliometrics.sciencedirect import ArticleRetrieval, init
from pybliometrics.sciencedirect.article_retrieval import Author

init()

ar_full = ArticleRetrieval('S2949948823000112', view='FULL', refresh=30)
ar_meta = ArticleRetrieval("10.1016/j.procs.2018.05.069", id_type='doi', view='META', refresh=30)
ar_meta_abs = ArticleRetrieval("S0304387823001736", view='META_ABS', refresh=30)
ar_meta_abs_ref = ArticleRetrieval("S2352226722000630", view='META_ABS_REF', refresh=30)
ar_entitled = ArticleRetrieval("S2213305418300365", view='ENTITLED', refresh=30)


def test_abstract():
    abstract_full = 'Artificial Neural Networks (ANNs) are a type of machine learning algorithm inspired by the structure and function of the human brain. In the context of supply chain management, ANNs can be used for demand forecasting, inventory optimization, logistics planning, and anomaly detection. ANNs help companies to optimize their inventory levels, production schedules and procurement activities in terms of productivity enhancement of part production. By considering multiple variables and constraints, ANNs can identify the most efficient routes, allocate resources effectively, and reduce costs. Furthermore, ANNs can identify anomalies as well as abnormalities in supply chain data, such as unexpected demand patterns, quality issues and disruptions in logistics operations in order to minimize their impact on the supply chain. ANNs can also analyze supplier performance data, including quality, delivery times and pricing in order to assess the reliability and effectiveness of suppliers. This information can support decision-making processes in supplier evaluation and selection processes. Moreover, ANNs can continuously monitor supplier performance, raising alerts for deviations from predefined criteria to provide safe and secure supply chain in part production processes. By analyzing various data sources, including weather conditions, and political instability, ANNs can identify and mitigate risks in terms of safety enhancement of supply chain processes. Artificial neural networks in supply chain management is studied in the research work to analyze and enhance performances of supply chain management in process of part manufacturing. New ideas and concepts of future research works are presented by reviewing and analyzing of recent achievements in applications of artificial neural networks in supply chain management. Thus, productivity of part manufacturing can be enhanced by promoting the supply chain management using the artificial neural networks.'
    abstract_meta_abs = 'This paper quantifies the relative footprint of trade and agricultural productivity on deforestation across Brazilian municipalities between 2000 and 2017. Using remote-sensing data, we identify distinct effects of these two phenomena on land use. Greater exposure to new genetically engineered soy seeds is associated with faster deforestation through cropland expansion. We find no significant association between local exposure to Chinese demand and deforestation, but exposure to trade with China mitigates the deforestation impacts from the new soy technology. Our findings suggest that, when considered together, productivity gains altering municipalities’ comparative advantage played a more significant role in driving deforestation across Brazil than Chinese demand alone.'
    abstract_meta_abs_ref = 'Subsidence along the central coast of the Ganges-Brahmaputra Delta and a subsequent rapid burial by mangrove mud have preserved a hundred earthen kilns at 11 localities, exposed due to recent coastal erosion. The associated pottery indicates that the kilns were used for salt crystallization in brine pots mounted on perforated hotplates. Four kiln varieties were found at successive elevations. The kiln chamber walls and the basal charcoal layers were dated using thermoluminescence and radiocarbon dating respectively. The oldest kilns with a round outline, found at the lowest elevations, were operating in the early 8th to middle 9th century. Oval-shaped kilns, found at a 1-m higher elevation, were used from the mid-9th to early 10th century. After a 600-year long period without documented kilns, rectangular kilns with a lateral slot were built again at a 1-m higher land surface in the 16th and 17th century. The youngest kiln type consists of rectangular twin chambers dating from the late 16th to mid-18th century and occurs at 1.5\xa0m below the present sediment surface. Architecture and functioning of these rectangular kilns are reconstructed, showing a multi-step procedure. This technique had been improved in the 19th century when pyramidal kilns were described in historic records.'
    assert ar_full.abstract == abstract_full
    assert ar_meta.abstract is None
    assert ar_meta_abs.abstract == abstract_meta_abs
    assert ar_meta_abs_ref.abstract == abstract_meta_abs_ref


def test_aggregationType():
    assert ar_full.aggregationType == 'Journal'
    assert ar_meta.aggregationType == 'Journal'
    assert ar_meta_abs.aggregationType == 'Journal'
    assert ar_meta_abs_ref.aggregationType == 'Journal'


def test_authors():
    authors_full = [Author(surname='Soori', given_name='Mohsen'),
                          Author(surname='Arezoo', given_name='Behrooz'),
                          Author(surname='Dastres', given_name='Roza')]
    authors_meta = [Author(surname='Indolia', given_name='Sakshi'),
                    Author(surname='Goswami', given_name='Anil Kumar'),
                    Author(surname='Mishra', given_name='S.P.'),
                    Author(surname='Asopa', given_name='Pooja')]
    authors_meta_abs = [Author(surname='Carreira', given_name='Igor'),
                        Author(surname='Costa', given_name='Francisco'),
                        Author(surname='Pessoa', given_name='João Paulo')]
    authors_meta_abs_ref = [Author(surname='Kudrass', given_name='H.R.'),
                            Author(surname='Hanebuth', given_name='T.J.J.'),
                            Author(surname='Zander', given_name='A.M.'),
                            Author(surname='Linstädter', given_name='J.'),
                            Author(surname='Akther', given_name='S.H.'),
                            Author(surname='Shohrab', given_name='U.M.')]
    assert ar_full.authors == authors_full
    assert ar_meta.authors == authors_meta
    assert ar_meta_abs.authors == authors_meta_abs
    assert ar_meta_abs_ref.authors == authors_meta_abs_ref


def test_copyright():
    copyright_full = '© 2024 The Author(s). Published by Elsevier B.V. on behalf of KeAi Communications Co., Ltd.'
    copyright_meta = '© 2018 The Author(s). Published by Elsevier B.V.'
    copyright_meta_abs = '© 2023 Elsevier B.V. All rights reserved.'
    copyright_meta_abs_ref = '© 2022 The Authors. Published by Elsevier Ltd.'
    assert ar_full.copyright == copyright_full
    assert ar_meta.copyright == copyright_meta
    assert ar_meta_abs.copyright == copyright_meta_abs
    assert ar_meta_abs_ref.copyright == copyright_meta_abs_ref


def test_coverDate():
    assert ar_full.coverDate == '2023-11-30'
    assert ar_meta.coverDate == '2018-12-31'
    assert ar_meta_abs.coverDate == '2024-03-31'
    assert ar_meta_abs_ref.coverDate == '2022-12-31'


def test_coverDisplayDate():
    assert ar_full.coverDisplayDate == 'November 2023'
    assert ar_meta.coverDisplayDate == '2018'
    assert ar_meta_abs.coverDisplayDate == 'March 2024'
    assert ar_meta_abs_ref.coverDisplayDate == 'December 2022'


def test_document_entitlement_status():
    assert ar_entitled.document_entitlement_status == 'ENTITLED'


def test_doi():
    assert ar_full.doi == '10.1016/j.ject.2023.11.002'
    assert ar_meta.doi == '10.1016/j.procs.2018.05.069'
    assert ar_meta_abs.doi == '10.1016/j.jdeveco.2023.103217'
    assert ar_meta_abs_ref.doi == '10.1016/j.ara.2022.100412'


def test_eid():
    assert ar_full.eid == '1-s2.0-S2949948823000112'
    assert ar_meta.eid == '1-s2.0-S1877050918308019'
    assert ar_meta_abs.eid == '1-s2.0-S0304387823001736'
    assert ar_meta_abs_ref.eid == '1-s2.0-S2352226722000630'


def test_endingPage():
    assert ar_full.endingPage == '196'
    assert ar_meta.endingPage == '688'
    assert ar_meta_abs.endingPage is None
    assert ar_meta_abs_ref.endingPage is None


def test_issn():
    assert ar_full.issn == 29499488
    assert ar_meta.issn == 18770509
    assert ar_meta_abs.issn == 3043878
    assert ar_meta_abs_ref.issn == 23522267


def test_openArchiveArticle():
    assert ar_full.openArchiveArticle is False
    assert ar_meta.openArchiveArticle is False
    assert ar_meta_abs.openArchiveArticle is False
    assert ar_meta_abs_ref.openArchiveArticle is False


def test_openaccess():
    assert ar_full.openaccess is True
    assert ar_meta.openaccess is True
    assert ar_meta_abs.openaccess is False
    assert ar_meta_abs_ref.openaccess is True


def test_openaccessSponsorName():
    assert ar_full.openaccessSponsorName == 'KeAi Communications Co., Ltd'
    assert ar_meta.openaccessSponsorName is None
    assert ar_meta_abs.openaccessSponsorName is None
    assert ar_meta_abs_ref.openaccessSponsorName is None


def test_openaccessSponsorType():
    assert ar_full.openaccessSponsorType == 'FundingBody'
    assert ar_meta.openaccessSponsorType == 'ElsevierWaived'
    assert ar_meta_abs.openaccessSponsorType is None
    assert ar_meta_abs_ref.openaccessSponsorType == 'Author'


def test_openaccessType():
    assert ar_full.openaccessType == 'Full'
    assert ar_meta.openaccessType == 'Full'
    assert ar_meta_abs.openaccessType is None
    assert ar_meta_abs_ref.openaccessType == 'Full'


def test_openaccessUserLicense():
    assert ar_full.openaccessUserLicense == 'http://creativecommons.org/licenses/by/4.0/'
    assert ar_meta.openaccessUserLicense == 'http://creativecommons.org/licenses/by-nc-nd/4.0/'
    assert ar_meta_abs.openaccessUserLicense is None
    assert ar_meta_abs_ref.openaccessUserLicense == 'http://creativecommons.org/licenses/by/4.0/'


def test_originalText():
    ar_full_test_originalText = 'c and Kuźnar, 2021). By analyzing data from multiple sources, including social media, news feeds, and sensor data, ANNs can provide early warnings and recommend proactive measures to mitigate potential risks. ANNs can be used to identify and mitigate supply chain risks by analyzing historical data and external factors (Liu, 2022). By training an ANN model with historical risk data and relevant variables, it can identify patterns and correlations which can be used in assessing the likelihood and impacts of potential risks, such as supplier disruptions, natural disasters and market volatility (Kosasih et al., 2022). This information allows organizations to take proactive measures to mitigate risks and improve supply chain resilience. ANNs can be used to analyze and mitigate risks at various stages of the supply chain, including procurement, production, transportation, and distribution during advanced supply chain management (Jianying et al., 2021). Here are some ways ANNs are used in supply chain risk analysis and mitigation: 1. Demand Forecasting: ANNs can be used to predict future demand patterns based on historical sales data, market trends, and other relevant factors. Accurate demand forecasting helps in inventory management and reduces the risk of stockouts or excess inventory (Aamer et al., 2020). 2. Supplier Evaluation: ANNs can assess supplier performance by analyzing various factors such as quality, delivery reliability, and pricing. By considering historical data and other relevant variables, ANNs can identify high-risk suppliers and help in making informed decisions about supplier selection and management (Hui and Choi, 2016). 3. Quality Control: ANNs can analyze quality-related data to identify patterns and anomalies that may indicate potential quality issues. By monitoring and analyzing data from production processes and quality inspections, ANNs can help in early detection and mitigation of quality-related risks (Cai et al., 2020; Minis, 2007). 4. Transp'
    assert ar_full.originalText[-80000:-78000] == ar_full_test_originalText


def test_pageRange():
    assert ar_full.pageRange == '179-196'
    assert ar_meta.pageRange == '679-688'
    assert ar_meta_abs.pageRange == '103217'
    assert ar_meta_abs_ref.pageRange == '100412'


def test_pii():
    assert ar_full.pii == 'S2949-9488(23)00011-2'
    assert ar_meta.pii == 'S1877-0509(18)30801-9'
    assert ar_meta_abs.pii == 'S0304-3878(23)00173-6'
    assert ar_meta_abs_ref.pii == 'S2352-2267(22)00063-0'


def test_pubType():
    assert ar_full.pubType == 'rev'
    assert ar_meta.pubType == 'fla'
    assert ar_meta_abs.pubType == 'fla'
    assert ar_meta_abs_ref.pubType == 'fla'


def test_publicationName():
    assert ar_full.publicationName == 'Journal of Economy and Technology'
    assert ar_meta.publicationName == 'Procedia Computer Science'
    assert ar_meta_abs.publicationName == 'Journal of Development Economics'
    assert ar_meta_abs_ref.publicationName == 'Archaeological Research in Asia'


def test_publisher():
    publisher_full = 'The Author(s). Published by Elsevier B.V. on behalf of KeAi Communications Co., Ltd.'
    assert ar_full.publisher == publisher_full
    assert ar_meta.publisher == 'The Author(s). Published by Elsevier B.V.'
    assert ar_meta_abs.publisher == 'Elsevier B.V.'
    assert ar_meta_abs_ref.publisher == 'The Authors. Published by Elsevier Ltd.'



def test_sciencedirect_link():
    assert ar_full.sciencedirect_link == 'https://www.sciencedirect.com/science/article/pii/S2949948823000112'
    assert ar_meta.sciencedirect_link == 'https://www.sciencedirect.com/science/article/pii/S1877050918308019'
    assert ar_meta_abs.sciencedirect_link == 'https://www.sciencedirect.com/science/article/pii/S0304387823001736'
    assert ar_meta_abs_ref.sciencedirect_link == 'https://www.sciencedirect.com/science/article/pii/S2352226722000630'


def test_self_link():
    assert ar_full.self_link == 'https://api.elsevier.com/content/article/pii/S2949948823000112'
    assert ar_meta.self_link == 'https://api.elsevier.com/content/article/pii/S1877050918308019'
    assert ar_meta_abs.self_link == 'https://api.elsevier.com/content/article/pii/S0304387823001736'
    assert ar_meta_abs_ref.sciencedirect_link == 'https://www.sciencedirect.com/science/article/pii/S2352226722000630'


def test_startingPage():
    assert ar_full.startingPage == '179'
    assert ar_meta.startingPage == '679'
    assert ar_meta_abs.startingPage == '103217'
    assert ar_meta_abs_ref.startingPage == '100412'


def test_subjects():
    assert ar_full.subjects == ['Artificial neural networks', 'Supply chain management']
    assert ar_meta.subjects == ['Convolution Neural Network', 'Deep Neural Network', 'Gradient Descent', 'ADAM']
    assert ar_meta_abs.subjects == ['Deforestation', 'Trade', 'Agricultural productivity', 'Technology', 'Soy', 'Brazil']
    assert ar_meta_abs_ref.subjects == ['Salt production', 'Earthen kiln', 'Ganges-Brahmaputra Delta', 'Sundarbans']


def test_title():
    assert ar_full.title == 'Artificial neural networks in supply chain management, a review'
    assert ar_meta.title == 'Conceptual Understanding of Convolutional Neural Network- A Deep Learning Approach '
    assert ar_meta_abs.title == 'The deforestation effects of trade and agricultural productivity in Brazil '
    assert ar_meta_abs_ref.title == 'Architecture and function of salt-producing kilns from the 8th to 18th century in the coastal Sundarbans mangrove forest, Central Ganges-Brahmaputra Delta, Bangladesh '


def test_url():
    assert ar_full.url == 'https://api.elsevier.com/content/article/pii/S2949948823000112'
    assert ar_meta.url == 'https://api.elsevier.com/content/article/pii/S1877050918308019'
    assert ar_meta_abs.url == 'https://api.elsevier.com/content/article/pii/S0304387823001736'
    assert ar_meta_abs_ref.url == 'https://api.elsevier.com/content/article/pii/S2352226722000630'


def test_volume():
    assert ar_full.volume == 1
    assert ar_meta.volume == 132
    assert ar_meta_abs.volume == 167
    assert ar_meta_abs_ref.volume == 32


def test__str__():
    expected  = 'Mohsen Soori, Behrooz Arezoo and Roza Dastres: "Artificial neural networks in supply chain management, a review", Journal of Economy and Technology, 1, pp. 179-196(2023). https://doi.org/10.1016/j.ject.2023.11.002.'
    assert str(ar_full) == expected



"""Tests for `sciencedirect.SubjectClassifications` module."""

from pybliometrics.sciencedirect import SubjectClassifications, init

init()

# Search by words in subject description
sub1 = SubjectClassifications({'description': 'Chemistry'}, refresh=30)
# Search by subject code
sub2 = SubjectClassifications({'code': '5'}, refresh=30)
# Search by words in subject detail
sub3 = SubjectClassifications({'detail': 'Optimization'}, refresh=30)
# Search by subject abbreviation
sub4 = SubjectClassifications({'abbrev': 'socialsciences'}, refresh=30)
# Search by multiple criteria
sub5 = SubjectClassifications({'description': 'Mathematics', 'detail':'Mathematics::Applied Mathematics'}, refresh=30)
# Search by multiple criteria, subset returned fields
sub6 = SubjectClassifications({'detail':'Agricultural and Biological Sciences', 'description': 'Agricultural and Biological Sciences'},
                              fields=['description', 'detail'], refresh=30)


def test_module():
    assert sub6.__class__.__name__  == 'ScDirSubjectClassifications'
    assert sub6.__module__  == 'pybliometrics.sciencedirect.subject_classifications'


def test_results_desc():
    assert len(sub1.results) > 0
    descriptions = [res.description for res in sub1.results]
    assert all([('Chemistry' in word or 'chemistry' in word) for word in descriptions])


def test_results_code():
    assert len(sub2.results) == 1
    assert sub2.results[0].code == '5'
    assert sub2.results[0].abbrev == 'agribio'
    

def test_results_detail():
    assert len(sub3.results) > 0
    assert all(['Optimization' in res.detail for res in sub3.results])


def test_results_abbrev():
    assert len(sub4.results) > 0
    assert all(['socialsciences' in res.abbrev for res in sub4.results])


def test_results_multi():
    assert len(sub5.results) > 0
    assert all(['Mathematics' in res.description for res in sub5.results])
    assert all(['Applied Mathematics' in res.detail for res in sub5.results])
    

def test_results_fields():
    assert len(sub6.results) > 0
    assert all(['Agricultural and Biological Sciences' in res.description for res in sub6.results])
    assert all(['Agricultural and Biological Sciences' in res.detail for res in sub6.results])
    assert all([set(res._fields) == set(['description', 'detail']) for res in sub6.results])



from typing import NamedTuple

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value, \
    get_id, get_link, parse_date_created, make_int_if_possible, VIEWS


class Variant(NamedTuple):
    name: str
    doc_count: int | None


class AffiliationRetrieval(Retrieval):
    @property
    def address(self) -> str | None:
        """The address of the affiliation."""
        return self._json.get('address')

    @property
    def affiliation_name(self) -> str:
        """The name of the affiliation."""
        return self._json.get('affiliation-name')

    @property
    def author_count(self) -> int:
        """Number of authors associated with the affiliation."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'author-count']))

    @property
    def city(self) -> str | None:
        """The city of the affiliation."""
        return self._json.get('city')

    @property
    def country(self) -> str | None:
        """The country of the affiliation."""
        return self._json.get('country')

    @property
    def date_created(self) -> tuple[int, int, int] | None:
        """Date the Scopus record was created."""
        try:
            return parse_date_created(self._profile)
        except KeyError:
            return None

    @property
    def document_count(self) -> int:
        """Number of documents for the affiliation."""
        return make_int_if_possible(chained_get(self._json, ['coredata', 'document-count']))
    
    @property
    def document_entitlement_status(self) -> str | None:
        """Returns the document entitlement status, i.e. tells if the requestor 
        is entitled to the requested resource.
        Note: Only works with `ENTITLED` view.
        """
        return chained_get(self._json, ['document-entitlement', 'status'])

    @property
    def eid(self) -> str:
        """The EID of the affiliation."""
        return chained_get(self._json, ['coredata', 'eid'])

    @property
    def identifier(self) -> int:
        """The Scopus ID of the affiliation."""
        return get_id(self._json)

    @property
    def name_variants(self) -> list[Variant] | None:
        """A list of namedtuples representing variants of the `affiliation_name`
        with number of documents referring to this variant.
        """
        path = ['name-variants', 'name-variant']
        variants = [Variant(name=var['$'],
                            doc_count=make_int_if_possible(var.get('@doc-count')))
                    for var in chained_get(self._json, path, [])]
        return variants or None

    @property
    def org_domain(self) -> str | None:
        """Internet domain of the affiliation.  Requires the STANDARD view."""
        return self._profile.get('org-domain')

    @property
    def org_type(self) -> str | None:
        """Type of the affiliation.  Requires the STANDARD view and only
        present if `profile` is `org profile`.
        """
        return self._profile.get('org-type')

    @property
    def org_URL(self) -> str | None:
        """Website of the affiliation.  Requires the STANDARD view."""
        return self._profile.get('org-URL')

    @property
    def postal_code(self) -> str | None:
        """The postal code of the affiliation.  Requires the STANDARD view."""
        return chained_get(self._profile, ['address', 'postal-code'])

    @property
    def scopus_affiliation_link(self) -> str:
        """Link to the Scopus web view of the affiliation."""
        return get_link(self._json, 2)

    @property
    def self_link(self) -> str:
        """Link to the affiliation's API page."""
        return get_link(self._json, 0)

    @property
    def search_link(self) -> str:
        """URL to the API page listing documents of the affiliation."""
        return get_link(self._json, 1)

    @property
    def state(self) -> str | None:
        """The state (country's administrative sububunit)
        of the affiliation.   Requires the STANDARD view.
        """
        return chained_get(self._profile, ['address', 'state'])

    @property
    def status(self) -> str | None:
        return self._profile.get("status")

    @property
    def sort_name(self) -> str | None:
        """The name of the affiliation used for sorting.  Requires the
        STANDARD view.
        """
        return self._profile.get('sort-name')

    @property
    def url(self) -> str:
        """URL to the affiliation's API page."""
        return chained_get(self._json, ['coredata', 'prism:url'])

    def __init__(self,
                 aff_id: int | str,
                 refresh: bool | int = False,
                 view: str = "STANDARD",
                 **kwds: str
                 ) -> None:
        """Interaction with the Affiliation Retrieval API.

        :param aff_id: Scopus ID or EID of the affiliation profile.
        :param refresh: Whether to refresh the cached file if it exists or not.
                        If int is passed, cached file will be refreshed if the
                        number of days since last modification exceeds that value.
        :param view: The view of the file that should be downloaded.  Allowed
                     values: `LIGHT`, `STANDARD`, `ENTITLED`, where `STANDARD` includes all
                     information of the `LIGHT` view.  For details see
                     https://dev.elsevier.com/sc_affil_retrieval_views.html.
                     Note: Neither the `BASIC` view nor `DOCUMENTS` or `AUTHORS`
                     views are active, although documented. `ENTITLED` only contains the `document_entitlement_status`.
        :param kwds: Keywords passed on as query parameters.  Must contain
                     fields and values mentioned in the API specification at
                     https://dev.elsevier.com/documentation/AffiliationRetrievalAPI.wadl.

        Raises
        ------
        ValueError
            If any of the parameters `refresh` or `view` is not
            one of the allowed values.

        Notes
        -----
        The directory for cached results is `{path}/{view}/{aff_id}`,
        where `path` is specified in your configuration file.
        """
        # Checks
        check_parameter_value(view, VIEWS['AffiliationRetrieval'], "view")

        # Load json
        self._view = view
        self._refresh = refresh
        aff_id = str(int(str(aff_id).split('-')[-1]))
        Retrieval.__init__(self, aff_id, **kwds)
        if self._view in ('LIGHT', 'STANDARD'):
            self._json = self._json['affiliation-retrieval-response']
        self._profile = self._json.get("institution-profile", {})

    def __str__(self):
        """Return a summary string."""
        date = self.get_cache_file_mdate().split()[0]
        s = f"{self.affiliation_name} in {self.city} in {self.country},\nhas "\
            f"{int(self.author_count):,} associated author(s) and "\
            f"{int(self.document_count):,} associated document(s) as of {date}"
        return s



"""Module for retrieving article entitlement information from ScienceDirect."""

from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import chained_get, check_parameter_value, detect_id_type, VIEWS


class ArticleEntitlement(Retrieval):
    """Class to retrieve the entitlement status for a document from ScienceDirect."""
    @property
    def status(self) -> str | None:
        """Status of whether a document has been found"""
        return self._json.get("@status")

    @property
    def identifier(self) -> str | None:
        """Identifier of a document."""
        return self._json.get("dc:identifier")

    @property
    def eid(self) -> str | None:
        "The EID of a document."
        return self._json.get("eid")

    @property
    def entitled(self) -> str | None:
        """Entitlement status of a document."""
        return self._json.get("entitled")

    @property
    def link(self) -> str | None:
        """ScienceDirect canonical URL."""
        return chained_get(self._json, ['link', '@href'])

    @property
    def message(self) -> str | None:
        """Entitlement status message."""
        return self._json.get("message")

    @property
    def pii(self) -> str | None:
        """The PII of a document."""
        return self._json.get("pii")

    @property
    def pii_norm(self) -> str | None:
        """The PII-norm of a document."""
        return self._json.get("pii-norm")

    @property
    def doi(self) -> str | None:
        """The DOI of a document."""
        return self._json.get("prism:doi")

    @property
    def pubmed_id(self) -> str | None:
        """The Pubmed ID of a document (when used in the request)."""
        return self._json.get("pubmed_id")

    @property
    def url(self) -> str | None:
        """API URL used to check entitlement."""
        return self._json.get("prism:url")

    @property
    def scopus_id(self) -> str | None:
        """The Scopus ID of a document (when used in the request)."""
        return self._json.get("scopus_id")

    def __init__(self,
                 identifier: int | str,
                 view: str = "FULL",
                 id_type: str | None = None,
                 refresh: bool | int = False,
                 **kwds: str) -> None:
        # Checks
        identifier = str(identifier)
        check_parameter_value(view, VIEWS["ArticleEntitlement"], "view")
        if not id_type:
            id_type = detect_id_type(identifier)
        else:
            allowed_id_types = ("eid", "pii", "scopus_id", "pubmed_id", "doi", "pui")
            check_parameter_value(id_type, allowed_id_types, "id_type")

        self._view = view
        self._refresh = refresh
        # Retrieve and get content
        Retrieval.__init__(self, identifier=identifier, id_type=id_type, **kwds)
        self._json = chained_get(self._json, ["entitlement-response", "document-entitlement"])

    def __str__(self) -> str:
        s = self.message
        s += f' with doi: {self.doi}'
        return s



"""Module to retrieve a specific object of a document."""

from io import BytesIO

from pybliometrics.sciencedirect import ArticleRetrieval
from pybliometrics.superclasses import Retrieval
from pybliometrics.utils import check_parameter_value, detect_id_type


class ObjectRetrieval(Retrieval):
    @property
    def object(self) -> BytesIO:
        """The object retrieved."""
        return BytesIO(self._object)

    def __init__(self,
                 identifier: int | str,
                 filename: str,
                 id_type: str | None = None,
                 refresh: bool | int = False,
                 **kwds: str
                 ):
        """Class to retrieve a specific object of a document by its filename.

        :param identifier: The indentifier of the document.
        :param filename: Filename of the object to be retrieved.  To get a list
                         of all available objects of a document (and its
                         corresponding filename) use the class `ObjectMetadata`.
        :param id_type: Document identifier.  Allowed values: `doi`, `pii`,
                        `scopus_id`, `pubmed_id`, `eid`.
        :param refresh: Whether to refresh the cached file if it exists.  Default: False.
        """
        identifier = str(identifier)

        if id_type is None:
            id_type = detect_id_type(identifier)
        else:
            allowed_id_types = ('doi', 'pii', 'scopus_id', 'pubmed_id', 'eid')
            check_parameter_value(id_type, allowed_id_types, "id_type")

        if id_type != 'eid':
            identifier = self._get_eid(identifier)
        file_identifier = f'{identifier}-{filename}'

        self._identifier = identifier
        self._filename = filename
        self._view = ''
        self._refresh = refresh

        super().__init__(file_identifier, 'eid', **kwds)

    def _get_eid(self, identifier: str) -> str:
        """Get the EID of a document."""
        am = ArticleRetrieval(identifier, field='eid')
        return am.eid

    def __str__(self) -> str:
        """Return a string with the filename, document and object size in KB."""
        size_kb = f"{len(self._object) / 1024:.1f}" if self._object else "0.0"
        return (f"Object {self._filename} from document with EID {self._identifier}"
                f" has size of {size_kb} KB.")
